import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.models import Message, Redditor, Subreddit, SubredditMessage

from ... import IntegrationTest


class TestMessage(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_attributes(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            messages = await self.async_list(self.reddit.inbox.messages())
            count = len(messages)
            while messages:
                message = messages.pop(0)
                messages.extend(message.replies)
                count -= 1
                try:
                    assert message.author is None or isinstance(
                        message.author, Redditor
                    )
                    assert isinstance(message.dest, (Redditor, Subreddit))
                    assert isinstance(message.replies, list)
                    assert message.subreddit is None or isinstance(
                        message.subreddit, Subreddit
                    )
                except Exception:
                    import pprint

                    pprint.pprint(vars(message))
                    raise
        assert count < 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_block(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            message = None
            async for item in self.reddit.inbox.messages():
                if item.author and item.author != pytest.placeholders.username:
                    message = item
                    break
            else:
                assert False, "no message found"
            await message.block()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            message = await self.async_next(self.reddit.inbox.messages())
            await message.delete()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mark_read(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            message = None
            async for item in self.reddit.inbox.unread():
                if isinstance(item, Message):
                    message = item
                    break
            else:
                assert False, "no message found in unread"
            await message.mark_read()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mark_unread(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            message = await self.async_next(self.reddit.inbox.messages())
            await message.mark_unread()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_message_collapse(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            message = await self.async_next(self.reddit.inbox.messages())
            await message.collapse()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_message_uncollapse(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            message = await self.async_next(self.reddit.inbox.messages())
            await message.uncollapse()

    async def test_parent(self):
        self.reddit.read_only = False
        with self.use_cassette():
            message = await self.reddit.inbox.message("1aogu6u")
            parent = message.parent
            assert isinstance(parent, Message)
            assert parent.fullname == message.parent_id

    async def test_parent__from_inbox_listing(self):
        self.reddit.read_only = False
        with self.use_cassette():
            message = await self.async_next(self.reddit.inbox.sent(limit=1))
            parent = message.parent
            assert isinstance(parent, Message)
            assert parent.fullname == message.parent_id
            assert not parent._fetched
            with pytest.raises(AttributeError) as excinfo:
                _ = parent.parent
                assert (
                    excinfo.value.args[0]
                    == "Message must be fetched with `.load()` before accessing the"
                    " parent."
                )
            await parent.load()
            assert isinstance(parent.parent, Message)
            assert parent.body

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_reply(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            message = await self.async_next(self.reddit.inbox.messages())
            reply = await message.reply("Message reply")
            assert reply.author == pytest.placeholders.username
            assert reply.body == "Message reply"
            assert reply.first_message_name == message.fullname

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unblock_subreddit(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            message1 = await self.async_next(self.reddit.inbox.messages(limit=1))
            assert isinstance(message1, SubredditMessage)
            message_fullname = message1.fullname
            await message1.block()
            message2 = await self.async_next(self.reddit.inbox.messages(limit=1))
            assert message2.fullname == message_fullname
            assert message2.subject == "[message from blocked subreddit]"
            await message2.unblock_subreddit()
            message3 = await self.async_next(self.reddit.inbox.messages(limit=1))
            assert message3.fullname == message_fullname
            assert message3.subject != "[message from blocked subreddit]"


class TestSubredditMessage(IntegrationTest):
    async def test_mute(self):
        self.reddit.read_only = False
        with self.use_cassette():
            message = SubredditMessage(self.reddit, _data={"id": "faj6z"})
            await message.mute()

    async def test_unmute(self):
        self.reddit.read_only = False
        with self.use_cassette():
            message = SubredditMessage(self.reddit, _data={"id": "faj6z"})
            await message.unmute()



"""Test praw.models.mod_notes."""

import pytest

from asyncpraw.models import Redditor, Subreddit

from .. import UnitTest


class TestBaseModNotes(UnitTest):
    async def test__ensure_attribute__(self):
        with pytest.raises(TypeError) as excinfo:
            Subreddit(None, "a").mod.notes._ensure_attribute(
                error_message="error", redditor=None
            )
        assert excinfo.value.args[0] == "error"

    async def test_notes_delete__missing_note_id(self):
        with pytest.raises(TypeError) as excinfo:
            await Subreddit(None, "a").mod.notes.delete(redditor="redditor")
        assert (
            excinfo.value.args[0]
            == "Either 'note_id' or 'delete_all' must be provided."
        )


class TestRedditModNotes(UnitTest):
    def test__call__invalid_thing_type(self, reddit):
        with pytest.raises(ValueError) as excinfo:
            reddit.notes(things=[1])
        assert (
            excinfo.value.args[0]
            == "Cannot get subreddit and author fields from type <class 'int'>"
        )

    def test__call__missing_arguments(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            reddit.notes()
        assert (
            excinfo.value.args[0]
            == "Either the 'pairs', 'redditors', 'subreddits', or 'things' parameters must be provided."
        )

    def test__call__redditors_missing_subreddits(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            reddit.notes(subreddits=[1])
        assert (
            excinfo.value.args[0]
            == "'redditors' must be non-empty if 'subreddits' is not empty."
        )


class TestRedditorModNotes(UnitTest):
    def test_subreddits__missing_argument(self):
        with pytest.raises(ValueError) as excinfo:
            Redditor(None, "a").notes.subreddits()
        assert excinfo.value.args[0] == "At least 1 subreddit must be provided."


class TestSubredditModNotes(UnitTest):
    def test_subreddits__missing_argument(self):
        with pytest.raises(ValueError) as excinfo:
            Subreddit(None, "a").mod.notes.redditors()
        assert excinfo.value.args[0] == "At least 1 redditor must be provided."



"""Test asyncpraw.models.util."""
from collections import namedtuple
from unittest import mock

import pytest

from asyncpraw.models.util import (
    BoundedSet,
    ExponentialCounter,
    permissions_string,
    stream_generator,
)

from .. import UnitTest


class TestExponentialCounter(UnitTest):
    MAX_DELTA = 1.0 / 32

    def test_exponential_counter__counter(self):
        def assert_range(number, exponent):
            assert number >= 2**exponent * (1 - self.MAX_DELTA)
            assert number <= 2**exponent * (1 + self.MAX_DELTA)

        counter = ExponentialCounter(1024)
        prev_value = counter.counter()
        assert_range(prev_value, 0)

        for i in range(9):
            value = counter.counter()
            assert_range(value, 1 + i)
            assert value > prev_value
            prev_value = value

    def test_exponential_counter__max_value(self):
        counter = ExponentialCounter(5)
        max_value = 5 * (1 + self.MAX_DELTA)
        for _ in range(100):
            value = counter.counter()
            assert value <= max_value

    def test_exponential_counter__reset(self):
        counter = ExponentialCounter(1024)
        for _ in range(100):
            value = counter.counter()
            assert value >= 1 - self.MAX_DELTA
            assert value <= 1 + self.MAX_DELTA
            counter.reset()


class TestBoundedSet(UnitTest):
    def test_contains(self):
        bset = BoundedSet(max_items=10)
        bset.add(1)
        assert 1 in bset

    def test_bound(self):
        bset = BoundedSet(max_items=10)
        [bset.add(i) for i in range(11)]
        assert len(bset._set) == 10
        assert 0 not in bset

    def test_lru_add(self):
        bset = BoundedSet(max_items=10)
        [bset.add(i) for i in range(10)]
        bset.add(0)
        bset.add(10)
        assert 0 in bset
        assert 1 not in bset

    def test_lru_contains(self):
        bset = BoundedSet(max_items=10)
        [bset.add(i) for i in range(10)]
        assert 0 in bset
        bset.add(10)
        assert 0 in bset
        assert 1 not in bset


class TestStream(UnitTest):
    @pytest.mark.asyncio
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_stream(self, _):
        Thing = namedtuple("Thing", ["fullname"])
        initial_things = [Thing(n) for n in reversed(range(100))]
        counter = 99

        async def generate(limit, **kwargs):
            nonlocal counter
            counter += 1
            if counter % 2 == 0:
                things = initial_things
            else:
                things = [Thing(counter)] + initial_things[:-1]
            for thing in things:
                yield thing

        stream = stream_generator(generate)
        seen = set()
        async for thing in stream:
            assert thing not in seen
            seen.add(thing)
            counter += 1
            if counter == 400:
                break


class TestPermissionsString(UnitTest):
    PERMISSIONS = {"a", "b", "c"}

    def test_permissions_string__all_explicit(self):
        assert "-all,+b,+a,+c" == permissions_string(
            known_permissions=self.PERMISSIONS, permissions=["b", "a", "c"]
        )

    def test_permissions_string__empty_list(self):
        assert "-all" == permissions_string(known_permissions=set(), permissions=[])
        assert "-all,-a,-b,-c" == permissions_string(
            known_permissions=self.PERMISSIONS, permissions=[]
        )

    def test_permissions_string__none(self):
        assert "+all" == permissions_string(known_permissions=set(), permissions=None)
        assert "+all" == permissions_string(
            known_permissions=self.PERMISSIONS, permissions=None
        )

    def test_permissions_string__with_additional_permissions(self):
        assert "-all,+d" == permissions_string(
            known_permissions=set(), permissions=["d"]
        )
        assert "-all,-a,-b,-c,+d" == permissions_string(
            known_permissions=self.PERMISSIONS, permissions=["d"]
        )



import pytest

from asyncpraw.exceptions import ClientException, RedditAPIException

from . import UnitTest


class TestObjector(UnitTest):
    def test_objectify_returns_None_for_None(self, reddit):
        assert reddit._objector.objectify(None) is None

    def test_parse_error(self, reddit):
        objector = reddit._objector
        assert objector.parse_error({}) is None
        assert objector.parse_error([]) is None
        assert objector.parse_error({"asdf": 1}) is None

        error_response = {
            "json": {"errors": [["USER_REQUIRED", "Please log in to do that.", None]]}
        }
        error = objector.parse_error(error_response)
        assert isinstance(error, RedditAPIException)

        error_response = {"json": {"errors": []}}
        with pytest.raises(ClientException):
            objector.parse_error(error_response)

        error_response = {
            "json": {
                "errors": [
                    ["USER_REQUIRED", "Please log in to do that.", None],
                    ["NO_SUBJECT", "please enter a subject", "subject"],
                ]
            }
        }
        assert isinstance(objector.parse_error(error_response), RedditAPIException)

    def test_check_error(self, reddit):
        objector = reddit._objector
        objector.check_error({"asdf": 1})

        error_response = {
            "json": {"errors": [["USER_REQUIRED", "Please log in to do that.", None]]}
        }
        with pytest.raises(RedditAPIException):
            objector.check_error(error_response)



"""Package providing models that pertain to listing mixins."""
from .base import BaseListingMixin
from .redditor import RedditorListingMixin
from .rising import RisingListingMixin
from .submission import SubmissionListingMixin
from .subreddit import SubredditListingMixin



"""Test asyncpraw.models.inbox."""
import sys

import pytest
from asyncprawcore import Forbidden

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.models import Comment, Message, Redditor, Subreddit

from .. import IntegrationTest


class TestInbox(IntegrationTest):
    async def test_all(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            async for item in self.reddit.inbox.all():
                assert isinstance(item, Comment) or isinstance(item, Message)
                count += 1
            assert count == 100

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_all__with_limit(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            assert len(await self.async_list(self.reddit.inbox.all(limit=128))) == 128

    async def test_comment_replies(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            async for item in self.reddit.inbox.comment_replies(limit=64):
                assert isinstance(item, Comment)
                assert item.parent_id.startswith(self.reddit.config.kinds["comment"])
                count += 1
            assert count == 64

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_comment_reply__refresh(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.async_next(self.reddit.inbox.comment_replies())
            saved_id = comment.id
            assert isinstance(comment, Comment)
            await comment.refresh()
            assert saved_id == comment.id

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mark_all_read(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await self.reddit.inbox.mark_unread(
                await self.async_list(self.reddit.inbox.all(limit=2))
            )
            await self.reddit.inbox.mark_all_read()
            assert not await self.async_list(self.reddit.inbox.unread())

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mark_read(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await self.reddit.inbox.mark_read(
                await self.async_list(self.reddit.inbox.unread())
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mark_unread(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await self.reddit.inbox.mark_unread(
                await self.async_list(self.reddit.inbox.all())
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mention__refresh(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            mention = await self.async_next(self.reddit.inbox.mentions())
            assert isinstance(mention, Comment)
            await mention.refresh()

    async def test_mentions(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            async for item in self.reddit.inbox.mentions(limit=16):
                assert isinstance(item, Comment)
                count += 1
            assert count > 0

    async def test_message(self):
        self.reddit.read_only = False
        with self.use_cassette():
            message = await self.reddit.inbox.message("kfrjvh")
        assert message.name.split("_", 1)[1] == "kfrjvh"
        assert isinstance(message, Message)
        assert isinstance(message.author, Redditor)
        assert isinstance(message.dest, Subreddit)
        assert message.replies == []
        assert isinstance(message.subreddit, Subreddit)

    async def test_message__unauthorized(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(Forbidden):
                await self.reddit.inbox.message("6i8om7")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_message_collapse(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await self.reddit.inbox.collapse(
                await self.async_list(self.reddit.inbox.messages())
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_message_uncollapse(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await self.reddit.inbox.uncollapse(
                await self.async_list(self.reddit.inbox.messages())
            )

    async def test_messages(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            async for item in self.reddit.inbox.messages(limit=64):
                assert isinstance(item, Message)
                count += 1
            assert count == 64

    async def test_sent(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            async for item in self.reddit.inbox.sent(limit=48):
                assert isinstance(item, Message)
                assert item.author == self.reddit.config.username
                count += 1
            assert count == 48

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_stream(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            item = await self.async_next(self.reddit.inbox.stream())
            assert isinstance(item, Comment) or isinstance(item, Message)

    async def test_submission_replies(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            async for item in self.reddit.inbox.submission_replies(limit=64):
                assert isinstance(item, Comment)
                assert item.parent_id.startswith(self.reddit.config.kinds["submission"])
                count += 1
            assert count == 64

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unread(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            async for item in self.reddit.inbox.unread(limit=64):
                count += 1
            assert count == 64



"""Represent the :class:`.Trophy` class."""
from typing import TYPE_CHECKING, Any, Dict, Union

from .base import AsyncPRAWBase

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class Trophy(AsyncPRAWBase):
    """Represent a trophy.

    End users should not instantiate this class directly. :meth:`.Redditor.trophies` can
    be used to get a list of the redditor's trophies.

    .. include:: ../../typical_attributes.rst

    =============== ===================================================
    Attribute       Description
    =============== ===================================================
    ``award_id``    The ID of the trophy (sometimes ``None``).
    ``description`` The description of the trophy (sometimes ``None``).
    ``icon_40``     The URL of a 41x41 px icon for the trophy.
    ``icon_70``     The URL of a 71x71 px icon for the trophy.
    ``name``        The name of the trophy.
    ``url``         A relevant URL (sometimes ``None``).
    =============== ===================================================

    """

    def __init__(self, reddit: "asyncpraw.Reddit", _data: Dict[str, Any]):
        """Initialize a :class:`.Trophy` instance.

        :param reddit: An instance of :class:`.Reddit`.
        :param _data: The structured data, assumed to be a dict and key ``"name"`` must
            be provided.

        """
        assert isinstance(_data, dict) and "name" in _data
        super().__init__(reddit, _data=_data)

    def __eq__(self, other: Union["Trophy", Any]) -> bool:
        """Check if two Trophies are equal."""
        if isinstance(other, self.__class__):
            return self.name == other.name
        return super().__eq__(other)

    def __str__(self) -> str:
        """Return a name of the trophy."""
        return self.name  # pylint: disable=no-member

    def __repr__(self) -> str:
        """Return an object initialization representation of the instance."""
        return f"{self.__class__.__name__}(name={self.name!r})"



"""Test asyncpraw.models.redditors."""

import sys

if sys.version_info < (3, 8):
    from asynctest import CoroutineMock as AsyncMock
    from asynctest import patch
else:
    from unittest.mock import patch, AsyncMock

from .. import UnitTest


class TestRedditors(UnitTest):
    async def test_search__params_not_modified(self, reddit):
        params = {"dummy": "value"}
        generator = reddit.redditors.search(None, params=params)
        assert generator.params["dummy"] == "value"
        assert params == {"dummy": "value"}

    async def test_partial_redditors(self, reddit):
        with patch.object(
            reddit, "request", AsyncMock(return_value={"data": {"children": []}})
        ) as mock_method:
            in_ids_list = [("t2_%d" % n) for n in range(100)]
            await self.async_list(reddit.redditors.partial_redditors(in_ids_list))

            assert mock_method.call_count == 1
            assert mock_method.call_args[1]["params"]["ids"] == ",".join(in_ids_list)

        with patch.object(
            reddit, "request", AsyncMock(return_value={"data": {"children": []}})
        ) as mock_method:
            in_ids_list = [("t2_%d" % n) for n in range(102)]
            await self.async_list(reddit.redditors.partial_redditors(in_ids_list))

            assert mock_method.call_count == 2
            cal = mock_method.call_args_list
            assert cal[0][1]["params"]["ids"] == ",".join(in_ids_list[:100])
            assert cal[1][1]["params"]["ids"] == ",".join(in_ids_list[-2:])

    async def test_partial_redditors__no_typeerror(self, reddit):
        func = reddit.redditors.partial_redditors
        with patch.object(reddit, "request"):
            func(list("abc"))
            func(tuple("abc"))
            func(c for c in "abc")



#!/usr/bin/env python3
"""This example demonstrates using the file token manager for refresh tokens.

In order to run this program, you will first need to obtain a valid refresh token. You
can use the ``obtain_refresh_token.py`` example to help.

In this example, refresh tokens will be saved into a file ``refresh_token.txt`` relative
to your current working directory. If your current working directory is under version
control it is strongly encouraged you add ``refresh_token.txt`` to the version control
ignore list.

Usage:

    EXPORT praw_client_id=<REDDIT_CLIENT_ID>
    EXPORT praw_client_secret=<REDDIT_CLIENT_SECRET>
    python3 use_file_token_manager.py

"""
import asyncio
import os
import sys

import aiofiles

import asyncpraw
from asyncpraw.util.token_manager import FileTokenManager

REFRESH_TOKEN_FILENAME = "refresh_token.txt"


async def initialize_refresh_token_file():
    if os.path.isfile(REFRESH_TOKEN_FILENAME):
        return

    refresh_token = input("Initial refresh token value: ")
    async with aiofiles.open(REFRESH_TOKEN_FILENAME, "w") as fp:
        await fp.write(refresh_token)


async def main():
    if "praw_client_id" not in os.environ:
        sys.stderr.write("Environment variable 'praw_client_id' must be defined\n")
        return 1
    if "praw_client_secret" not in os.environ:
        sys.stderr.write("Environment variable 'praw_client_secret' must be defined\n")
        return 1

    await initialize_refresh_token_file()

    refresh_token_manager = FileTokenManager(REFRESH_TOKEN_FILENAME)
    async with asyncpraw.Reddit(
        token_manager=refresh_token_manager,
        user_agent="use_file_token_manager/v0 by u/bboe",
    ) as reddit:

        scopes = await reddit.auth.scopes()
        if scopes == {"*"}:
            print(f"{await reddit.user.me()} is authenticated with all scopes")
        elif "identity" in scopes:
            print(
                f"{await reddit.user.me()} is authenticated with the following scopes:"
                f" {scopes}"
            )
        else:
            print(f"You are authenticated with the following scopes: {scopes}")


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    sys.exit(loop.run_until_complete(main()))



#!/usr/bin/env python3
import asyncio
from urllib.parse import quote_plus

import asyncpraw

QUESTIONS = ["what is", "who is", "what are"]
REPLY_TEMPLATE = "[Let me google that for you](https://lmgtfy.com/?q={})"


async def main():
    reddit = asyncpraw.Reddit(
        client_id="CLIENT_ID",
        client_secret="CLIENT_SECRET",
        password="PASSWORD",
        user_agent="LMGTFY (by u/USERNAME)",
        username="USERNAME",
    )

    subreddit = await reddit.subreddit("AskReddit")
    async for submission in subreddit.stream.submissions():
        await process_submission(submission)


async def process_submission(submission):
    # Ignore titles with more than 10 words as they probably are not simple
    # questions.
    if len(submission.title.split()) > 10:
        return

    normalized_title = submission.title.lower()
    for question_phrase in QUESTIONS:
        if question_phrase in normalized_title:
            url_title = quote_plus(submission.title)
            reply_text = REPLY_TEMPLATE.format(url_title)
            print(f"Replying to: {submission.title}")
            await submission.reply(reply_text)
            # A reply has been made so do not attempt to match other phrases.
            break


if __name__ == "__main__":
    asyncio.run(main())



"""Test asyncpraw.models.front."""
from .. import IntegrationTest


class TestFront(IntegrationTest):
    async def test_best(self):
        with self.use_cassette():
            submissions = await self.async_list(self.reddit.front.best())
        assert len(submissions) == 100

    async def test_controversial(self):
        with self.use_cassette():
            submissions = await self.async_list(self.reddit.front.controversial())
        assert len(submissions) == 100

    async def test_gilded(self):
        with self.use_cassette():
            submissions = await self.async_list(self.reddit.front.gilded())
        assert len(submissions) == 100

    async def test_hot(self):
        with self.use_cassette():
            submissions = await self.async_list(self.reddit.front.hot())
        assert len(submissions) == 100

    async def test_new(self):
        with self.use_cassette():
            submissions = await self.async_list(self.reddit.front.new())
        assert len(submissions) == 100

    async def test_top(self):
        with self.use_cassette():
            submissions = await self.async_list(self.reddit.front.top())
        assert len(submissions) == 100



"""Test asyncpraw.models.auth."""
import pytest

from asyncpraw import Reddit
from asyncpraw.exceptions import ClientException

from .. import UnitTest


class TestAuth(UnitTest):
    @pytest.fixture
    async def installed_app(self):
        async with Reddit(
            client_id="dummy client",
            client_secret=None,
            redirect_uri="https://dummy.tld/",
            user_agent="dummy",
        ) as reddit:
            yield reddit

    @pytest.fixture
    async def script_app(self):
        async with Reddit(
            client_id="dummy client",
            client_secret="dummy secret",
            redirect_uri="https://dummy.tld/",
            user_agent="dummy",
        ) as reddit:
            yield reddit

    @pytest.fixture
    async def script_app_with_password(self):
        async with Reddit(
            client_id="dummy client",
            client_secret="dummy secret",
            password="dummy password",
            user_agent="dummy",
            username="dummy username",
        ) as reddit:
            yield reddit

    @pytest.fixture
    async def web_app(self):
        async with Reddit(
            client_id="dummy client",
            client_secret="dummy secret",
            redirect_uri="https://dummy.tld/",
            user_agent="dummy",
        ) as reddit:
            yield reddit

    def test_implicit__from_script_app(self, script_app_with_password, script_app):
        with pytest.raises(ClientException):
            script_app.auth.implicit(
                access_token="dummy token", expires_in=10, scope=""
            )
        with pytest.raises(ClientException):
            script_app_with_password.auth.implicit(
                access_token="dummy token", expires_in=10, scope=""
            )

    def test_implicit__from_web_app(self, web_app):
        with pytest.raises(ClientException):
            web_app.auth.implicit(access_token="dummy token", expires_in=10, scope="")

    def test_limits(self, web_app, script_app_with_password, script_app, installed_app):
        expected = {"remaining": None, "reset_timestamp": None, "used": None}
        for app in [
            installed_app,
            script_app,
            script_app_with_password,
            web_app,
        ]:
            assert expected == app.auth.limits

    async def test_url__installed_app(self, installed_app):
        url = installed_app.auth.url(scopes=["dummy scope"], state="dummy state")
        assert "client_id=dummy+client" in url
        assert "duration=permanent" in url
        assert "redirect_uri=https://dummy.tld/" in url
        assert "response_type=code" in url
        assert "scope=dummy+scope" in url
        assert "state=dummy+state" in url

    async def test_url__installed_app__implicit(self, installed_app):
        url = installed_app.auth.url(
            implicit=True, scopes=["dummy scope"], state="dummy state"
        )
        assert "client_id=dummy+client" in url
        assert "duration=temporary" in url
        assert "redirect_uri=https://dummy.tld/" in url
        assert "response_type=token" in url
        assert "scope=dummy+scope" in url
        assert "state=dummy+state" in url

    def test_url__web_app(self, web_app):
        url = web_app.auth.url(scopes=["dummy scope"], state="dummy state")
        assert "client_id=dummy+client" in url
        assert "secret" not in url
        assert "redirect_uri=https://dummy.tld/" in url
        assert "response_type=code" in url
        assert "scope=dummy+scope" in url
        assert "state=dummy+state" in url

    def test_url__web_app__implicit(self, web_app):
        with pytest.raises(ClientException):
            web_app.auth.url(implicit=True, scopes=["dummy scope"], state="dummy state")

    async def test_url__web_app_without_redirect_uri(self):
        async with Reddit(
            client_id="dummy client", client_secret="dummy secret", user_agent="dummy"
        ) as reddit:
            with pytest.raises(ClientException):
                reddit.auth.url(scopes=["dummy scope"], state="dummy state")



"""Provide the MoreComments class."""
from typing import TYPE_CHECKING, Any, Dict, List, Union

from ...const import API_PATH
from ...util import _deprecate_args
from ..base import AsyncPRAWBase

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class MoreComments(AsyncPRAWBase):
    """A class indicating there are more comments."""

    def __init__(self, reddit: "asyncpraw.Reddit", _data: Dict[str, Any]):
        """Initialize a :class:`.MoreComments` instance."""
        self.count = self.parent_id = None
        self.children = []
        super().__init__(reddit, _data=_data)
        self._comments = None
        self.submission = None

    def __eq__(self, other: Union[str, "MoreComments"]) -> bool:
        """Return ``True`` if these :class:`.MoreComments` instances are the same."""
        if isinstance(other, self.__class__):
            return self.count == other.count and self.children == other.children
        return super().__eq__(other)

    def __lt__(self, other: "MoreComments") -> bool:
        """Provide a sort order on the :class:`.MoreComments` object."""
        # To work with heapq a "smaller" item is the one with the most comments. We are
        # intentionally making the biggest element the smallest element to turn the
        # min-heap implementation in heapq into a max-heap.
        return self.count > other.count

    def __repr__(self) -> str:
        """Return an object initialization representation of the instance."""
        children = self.children[:4]
        if len(self.children) > 4:
            children[-1] = "..."
        return f"<{self.__class__.__name__} count={self.count}, children={children!r}>"

    async def _continue_comments(self, update):
        assert not self.children, "Please file a bug report with Async PRAW."
        parent = await self._load_comment(self.parent_id.split("_", 1)[1])
        self._comments = parent.replies
        if update:
            for comment in self._comments:
                comment.submission = self.submission
        return self._comments

    async def _load_comment(self, comment_id):
        path = f"{API_PATH['submission'].format(id=self.submission.id)}_/{comment_id}"
        _, comments = await self._reddit.get(
            path,
            params={
                "limit": self.submission.comment_limit,
                "sort": self.submission.comment_sort,
            },
        )
        assert len(comments.children) == 1, "Please file a bug report with Async PRAW."
        return comments.children[0]

    @_deprecate_args("update")
    async def comments(
        self, *, update: bool = True
    ) -> List["asyncpraw.models.Comment"]:
        """Fetch and return the comments for a single :class:`.MoreComments` object."""
        if self._comments is None:
            if self.count == 0:  # Handle "continue this thread"
                return await self._continue_comments(update)
            assert self.children, "Please file a bug report with Async PRAW."
            data = {
                "children": ",".join(self.children),
                "link_id": self.submission.fullname,
                "sort": self.submission.comment_sort,
            }
            self._comments = await self._reddit.post(
                API_PATH["morechildren"], data=data
            )
            if update:
                for comment in self._comments:
                    comment.submission = self.submission
        return self._comments



"""Test asyncpraw.models.list.base."""
import pytest

from asyncpraw.models.list.base import BaseList

from ... import UnitTest


class DummyObjector:
    @staticmethod
    def objectify(value):
        return value


class Dummy:
    def __init__(self):
        self._objector = DummyObjector


class TestBaseList(UnitTest):
    def setup_method(self, _):
        self._prev_child_attribute = BaseList.CHILD_ATTRIBUTE

    def teardown_method(self, _):
        BaseList.CHILD_ATTRIBUTE = self._prev_child_attribute

    def test__init__CHILD_ATTRIBUTE_not_set(self):
        with pytest.raises(NotImplementedError):
            BaseList(None, None)

    def test__contains__(self):
        BaseList.CHILD_ATTRIBUTE = "asyncpraw"
        items = ["foo", 1, {"a": "b"}]
        base_list = BaseList(Dummy(), {"asyncpraw": items})
        for item in items:
            assert item in base_list

    def test__getitem__(self):
        BaseList.CHILD_ATTRIBUTE = "asyncpraw"
        items = ["foo", 1, {"a": "b"}]
        base_list = BaseList(Dummy(), {"asyncpraw": items})
        for i, item in enumerate(items):
            assert item == base_list[i]

    def test__iter__(self):
        BaseList.CHILD_ATTRIBUTE = "asyncpraw"
        items = ["foo", 1, {"a": "b"}]
        base_list = BaseList(Dummy(), {"asyncpraw": items})
        for i, item in enumerate(base_list):
            assert items[i] == item

    def test__str__(self):
        BaseList.CHILD_ATTRIBUTE = "asyncpraw"
        items = ["foo", 1, {"a": "b"}]
        base_list = BaseList(Dummy(), {"asyncpraw": items})
        assert str(items) == str(base_list)



"""Provide the Listing class."""
from typing import Any, Optional

from ..base import AsyncPRAWBase


class Listing(AsyncPRAWBase):
    """A listing is a collection of :class:`.RedditBase` instances."""

    CHILD_ATTRIBUTE = "children"

    def __len__(self) -> int:
        """Return the number of items in the Listing."""
        return len(getattr(self, self.CHILD_ATTRIBUTE))

    def __getitem__(self, index: int) -> Any:
        """Return the item at position index in the list."""
        return getattr(self, self.CHILD_ATTRIBUTE)[index]

    def __setattr__(self, attribute: str, value: Any):
        """Objectify the ``CHILD_ATTRIBUTE`` attribute."""
        if attribute == self.CHILD_ATTRIBUTE:
            value = self._reddit._objector.objectify(value)
        super().__setattr__(attribute, value)


class FlairListing(Listing):
    """Special Listing for handling flair lists."""

    CHILD_ATTRIBUTE = "users"

    @property
    def after(self) -> Optional[Any]:
        """Return the next attribute or ``None``."""
        return getattr(self, "next", None)


class ModeratorListing(Listing):
    """Special Listing for handling moderator lists."""

    CHILD_ATTRIBUTE = "moderators"


class ModNoteListing(Listing):
    """Special Listing for handling :class:`.ModNote` lists."""

    CHILD_ATTRIBUTE = "mod_notes"

    @property
    def after(self) -> Optional[Any]:
        """Return the next attribute or None."""
        if not getattr(self, "has_next_page", True):
            return None
        return getattr(self, "end_cursor", None)


class ModmailConversationsListing(Listing):
    """Special Listing for handling :class:`.ModmailConversation` lists."""

    CHILD_ATTRIBUTE = "conversations"

    @property
    def after(self) -> Optional[str]:
        """Return the next attribute or ``None``."""
        try:
            return self.conversations[-1].id
        except IndexError:
            return None



"""Test asyncpraw.models.subreddit."""
import socket
import sys
from asyncio import TimeoutError
from os.path import abspath, dirname, join

import pytest
from aiohttp import ClientResponse
from aiohttp.http_websocket import WebSocketError
from asyncprawcore import BadRequest, Forbidden, NotFound, TooLarge

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.const import PNG_HEADER
from asyncpraw.exceptions import (
    ClientException,
    InvalidFlairTemplateID,
    RedditAPIException,
    TooLargeMediaException,
    WebSocketException,
)
from asyncpraw.models import (
    Comment,
    InlineGif,
    InlineImage,
    InlineVideo,
    ListingGenerator,
    ModAction,
    ModmailAction,
    ModmailConversation,
    ModmailMessage,
    Redditor,
    Stylesheet,
    Submission,
    Subreddit,
    SubredditMessage,
    WikiPage,
)

from ... import IntegrationTest


def image_path(name):
    test_dir = abspath(dirname(sys.modules[__name__].__file__))
    return abspath(join(test_dir, "..", "..", "files", name))


class WebsocketMock:
    POST_URL = "https://reddit.com/r/<TEST_SUBREDDIT>/comments/{}/test_title/"

    @classmethod
    def make_dict(cls, post_id):
        return {"payload": {"redirect": cls.POST_URL.format(post_id)}}

    def __init__(self, *post_ids):
        self.post_ids = post_ids
        self.i = -1

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def receive_json(self):
        if not self.post_ids:
            raise WebSocketError(None, None)
        assert 0 <= self.i + 1 < len(self.post_ids)
        self.i += 1
        return self.make_dict(self.post_ids[self.i])


class TestSubreddit(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create(self, _):
        self.reddit.read_only = False
        new_name = pytest.placeholders.test_subreddit
        with self.use_cassette():
            subreddit = await self.reddit.subreddit.create(
                new_name,
                link_type="any",
                subreddit_type="public",
                title="Sub",
                wikimode="disabled",
                wiki_edit_age=0,
                wiki_edit_karma=0,
                comment_score_hide_mins=0,
            )
            assert subreddit.display_name == new_name
            assert subreddit.submission_type == "any"

    async def test_create__exists(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await self.reddit.subreddit.create(
                    "redditdev",
                    link_type="any",
                    subreddit_type="public",
                    title="redditdev",
                    wikimode="disabled",
                )
            assert excinfo.value.items[0].error_type == "SUBREDDIT_EXISTS"

    async def test_create__invalid_parameter(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                # Supplying invalid setting for link_type
                await self.reddit.subreddit.create(
                    "PRAW_iavynavff2",
                    link_type="abcd",
                    subreddit_type="public",
                    title="sub",
                    wikimode="disabled",
                )
            assert excinfo.value.items[0].error_type == "INVALID_OPTION"

    async def test_create__missing_parameter(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                # Not supplying required field wiki_edit_age.
                await self.reddit.subreddit.create(
                    "PRAW_iavynavff3",
                    link_type="any",
                    subreddit_type="public",
                    title=None,
                    wikimode="disabled",
                    wiki_edit_karma=0,
                    comment_score_hide_mins=0,
                )
            assert excinfo.value.items[0].error_type == "BAD_NUMBER"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_message(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.message(
                subject="Test from Async PRAW", message="Test content"
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_post_requirements(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            data = await subreddit.post_requirements()
            tags = [
                "domain_blacklist",
                "body_restriction_policy",
                "domain_whitelist",
                "title_regexes",
                "body_blacklisted_strings",
                "body_required_strings",
                "title_text_min_length",
                "is_flair_required",
                "title_text_max_length",
                "body_regexes",
                "link_repost_age",
                "body_text_min_length",
                "link_restriction_policy",
                "body_text_max_length",
                "title_required_strings",
                "title_blacklisted_strings",
                "guidelines_text",
                "guidelines_display_policy",
            ]
            assert list(data) == tags

    async def test_random(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("pics")
            submissions = [
                await subreddit.random(),
                await subreddit.random(),
                await subreddit.random(),
                await subreddit.random(),
            ]
        assert len(submissions) == len(set(submissions))

    async def test_random__returns_none(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("wallpapers")
            assert await subreddit.random() is None

    async def test_sticky(self):
        subreddit = await self.reddit.subreddit("pics")
        with self.use_cassette():
            submission = await subreddit.sticky()
            assert isinstance(submission, Submission)

    async def test_sticky__not_set(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            with pytest.raises(NotFound):
                await subreddit.sticky(number=2)

    async def test_search(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("all")
            async for item in subreddit.search(
                "praw oauth search", syntax="cloudsearch", limit=None
            ):
                assert isinstance(item, Submission)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit__flair(self, _):
        flair_id = "94f13282-e2e8-11e8-8291-0eae4e167256"
        flair_text = "AF"
        flair_class = ""
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit(
                "Test Title",
                flair_id=flair_id,
                flair_text=flair_text,
                selftext="Test text.",
            )
            await submission.load()
            assert submission.link_flair_css_class == flair_class
            assert submission.link_flair_text == flair_text

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit__selftext(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit("Test Title", selftext="Test text.")
            await submission.load()
            assert submission.author == pytest.placeholders.username
            assert submission.selftext == "Test text."
            assert submission.title == "Test Title"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit__selftext_blank(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit("Test Title", selftext="")
            await submission.load()
            assert submission.author == pytest.placeholders.username
            assert submission.selftext == ""
            assert submission.title == "Test Title"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit__selftext_inline_media(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            gif = InlineGif(image_path("test.gif"), "optional caption")
            image = InlineImage(image_path("test.png"), "optional caption")
            video = InlineVideo(image_path("test.mp4"), "optional caption")
            selftext = (
                "Text with a gif {gif1} an image {image1} and a video {video1} inline"
            )
            media = {"gif1": gif, "image1": image, "video1": video}
            submission = await subreddit.submit(
                "title", inline_media=media, selftext=selftext
            )
            await submission.load()
            assert submission.author == pytest.placeholders.username
            assert (
                submission.selftext == "Text with a gif\n\n[optional"
                " caption](https://i.redd.it/s1i7ejqkgdc61.gif)\n\nan"
                " image\n\n[optional"
                " caption](https://preview.redd.it/95pza2skgdc61.png?width=128&format=png&auto=webp&s=c81d303645d9792afcdb9c47f0a6039708714274)\n\nand"
                " a video\n\n[optional"
                " caption](https://reddit.com/link/l0vyxc/video/qeg2azskgdc61/player)\n\ninline"
            )
            assert submission.title == "title"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_live_chat(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit(
                "Test Title", discussion_type="CHAT", selftext=""
            )
            await submission.load()
            assert submission.discussion_type == "CHAT"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit__url(self, _):
        url = "https://asyncpraw.readthedocs.org/en/stable/"
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit("Test Title", url=url)
            await submission.load()
            assert submission.author == pytest.placeholders.username
            assert submission.url == url
            assert submission.title == "Test Title"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit__nsfw(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit(
                "Test Title", nsfw=True, selftext="Test text."
            )
            await submission.load()
            assert submission.over_18 is True

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit__spoiler(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit(
                "Test Title", selftext="Test text.", spoiler=True
            )
            await submission.load()
            assert submission.spoiler is True

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit__verify_invalid(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            self.reddit.validate_on_submit = True
            with pytest.raises(
                (RedditAPIException, BadRequest)
            ):  # waiting for prawcore fix
                await subreddit.submit("dfgfdgfdgdf", url="https://www.google.com")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_poll(self, _):
        options = ["Yes", "No", "3", "4", "5", "6"]
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit_poll(
                "Test Poll", duration=6, options=options, selftext="Test poll text."
            )
            await submission.load()
            assert submission.author == pytest.placeholders.username
            assert submission.selftext.startswith("Test poll text.")
            assert submission.title == "Test Poll"
            assert [str(option) for option in submission.poll_data.options] == options
            assert submission.poll_data.voting_end_timestamp > submission.created_utc

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_poll__flair(self, _):
        flair_id = "94f13282-e2e8-11e8-8291-0eae4e167256"
        flair_text = "AF"
        flair_class = ""
        options = ["Yes", "No"]
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit_poll(
                "Test Poll",
                duration=6,
                flair_id=flair_id,
                flair_text=flair_text,
                options=options,
                selftext="Test poll text.",
            )
            await submission.load()
            assert submission.link_flair_text == flair_text
            assert submission.link_flair_css_class == flair_class

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_poll__live_chat(self, _):
        options = ["Yes", "No"]
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit_poll(
                "Test Poll",
                discussion_type="CHAT",
                duration=2,
                options=options,
                selftext="",
            )
            await submission.load()
            assert submission.discussion_type == "CHAT"

    async def test_submit_gallery(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            images = [
                {"image_path": image_path("test.png")},
                {"image_path": image_path("test.jpg"), "caption": "test.jpg"},
                {
                    "image_path": image_path("test.gif"),
                    "outbound_url": "https://example.com",
                },
                {
                    "image_path": image_path("test.png"),
                    "caption": "test.png",
                    "outbound_url": "https://example.com",
                },
            ]

            submission = await subreddit.submit_gallery("Test Title", images)
            assert submission.author == pytest.placeholders.username
            assert submission.is_gallery
            assert submission.title == "Test Title"
            items = submission.gallery_data["items"]
            assert isinstance(submission.gallery_data["items"], list)
            for i, item in enumerate(items):
                test_data = images[i]
                test_data.pop("image_path")
                item.pop("id")
                item.pop("media_id")
                assert item == test_data

    async def test_submit_gallery_disabled(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            images = [
                {"image_path": image_path("test.png")},
                {"image_path": image_path("test.jpg"), "caption": "test.jpg"},
                {
                    "image_path": image_path("test.gif"),
                    "outbound_url": "https://example.com",
                },
                {
                    "image_path": image_path("test.png"),
                    "caption": "test.png",
                    "outbound_url": "https://example.com",
                },
            ]

            with pytest.raises(RedditAPIException):
                await subreddit.submit_gallery("Test Title", images)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_gallery__flair(self, _):
        flair_id = "6fc213da-cae7-11ea-9274-0e2407099e45"
        flair_text = "test"
        flair_class = "test-flair-class"
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            images = [
                {"image_path": image_path("test.png")},
                {"image_path": image_path("test.jpg"), "caption": "test.jpg"},
                {
                    "image_path": image_path("test.gif"),
                    "outbound_url": "https://example.com",
                },
                {
                    "image_path": image_path("test.png"),
                    "caption": "test.png",
                    "outbound_url": "https://example.com",
                },
            ]
            submission = await subreddit.submit_gallery(
                "Test Title", images, flair_id=flair_id, flair_text=flair_text
            )
            assert submission.link_flair_css_class == flair_class
            assert submission.link_flair_text == flair_text

    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        return_value=WebsocketMock(
            "l6eqw6", "l6er3r", "l6erfu"  # update with cassette
        ),
    )
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_image(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for i, file_name in enumerate(("test.png", "test.jpg", "test.gif")):
                image = image_path(file_name)
                submission = await subreddit.submit_image(f"Test Title {i}", image)
                await submission.load()
                assert submission.author == pytest.placeholders.username
                assert submission.is_reddit_media_domain
                assert submission.title == f"Test Title {i}"

    @pytest.mark.usefixtures("tmp_path")
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_image__large(self, _):
        reddit = self.reddit
        reddit.read_only = False
        mock_data = (
            '<?xml version="1.0" encoding="UTF-8"?>'
            "<Error>"
            "<Code>EntityTooLarge</Code>"
            "<Message>Your proposed upload exceeds the maximum allowed size</Message>"
            "<ProposedSize>20971528</ProposedSize>"
            "<MaxSizeAllowed>20971520</MaxSizeAllowed>"
            "<RequestId>23F056D6990D87E0</RequestId>"
            "<HostId>iYEVOuRfbLiKwMgHt2ewqQRIm0NWL79uiC2rPLj9P0PwW55MhjY2/O8d9JdKTf1iwzLjwWMnGQ=</HostId>"
            "</Error>"
        )
        _post = reddit._core._requestor._http.post

        async def patch_request(url, *args, **kwargs):
            """Patch requests to return mock data on specific url."""
            if "https://reddit-uploaded-media.s3-accelerate.amazonaws.com" in url:
                response = ClientResponse
                response.text = mock_data
                response.status = 400
                return response
            return await _post(url, *args, **kwargs)

        reddit._core._requestor._http.post = patch_request

        fake_png = PNG_HEADER + b"\x1a" * 10  # Normally 1024 ** 2 * 20 (20 MB)
        with open(self.tmp_path.joinpath("fake_img.png"), "wb") as tempfile:
            tempfile.write(fake_png)
        with self.use_cassette():
            with pytest.raises(TooLargeMediaException):
                subreddit = await reddit.subreddit("test")
                await subreddit.submit_image("test", tempfile.name)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch("aiohttp.client.ClientSession.ws_connect", return_value=WebsocketMock())
    async def test_submit_image__bad_websocket(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette("TestSubreddit.test_submit_image"):
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for file_name in ("test.png", "test.jpg"):
                image = image_path(file_name)
                with pytest.raises(ClientException):
                    await subreddit.submit_image("Test Title", image)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_image__bad_filetype(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for file_name in ("test.mov", "test.mp4"):
                image = image_path(file_name)
                with pytest.raises(ClientException):
                    await subreddit.submit_image("Test Title", image)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect", return_value=WebsocketMock("l6evpd")
    )  # update with cassette
    async def test_submit_image__flair(self, _, __):
        flair_id = "6fc213da-cae7-11ea-9274-0e2407099e45"
        flair_text = "Test flair text"
        flair_class = ""
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            image = image_path("test.jpg")
            submission = await subreddit.submit_image(
                "Test Title", image, flair_id=flair_id, flair_text=flair_text
            )
            await submission.load()
            assert submission.link_flair_css_class == flair_class
            assert submission.link_flair_text == flair_text

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect", return_value=WebsocketMock("l6ey7j")
    )  # update with cassette
    async def test_submit_image_chat(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            image = image_path("test.jpg")
            submission = await subreddit.submit_image(
                "Test Title", image, discussion_type="CHAT"
            )
            await submission.load()
            assert submission.discussion_type == "CHAT"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_image_verify_invalid(self, _):
        self.reddit.read_only = False
        self.reddit.validate_on_submit = True
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            image = image_path("test.jpg")
            with pytest.raises(
                (RedditAPIException, BadRequest)
            ):  # waiting for prawcore fix
                await subreddit.submit_image(
                    "gdfgfdgdgdgfgfdgdfgfdgfdg", image, without_websockets=True
                )

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect", side_effect=BlockingIOError
    )  # happens with timeout=0
    async def test_submit_image__timeout_1(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            image = image_path("test.jpg")
            with pytest.raises(WebSocketException):
                await subreddit.submit_image("Test Title", image)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        side_effect=socket.timeout
        # happens with timeout=0.00001
    )
    async def test_submit_image__timeout_2(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            image = image_path("test.jpg")
            with pytest.raises(WebSocketException):
                await subreddit.submit_image("Test Title", image)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        side_effect=TimeoutError,
        # could happen but Async PRAW should handle it
    )
    async def test_submit_image__timeout_3(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            image = image_path("test.jpg")
            with pytest.raises(WebSocketException):
                await subreddit.submit_image("Test Title", image)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        side_effect=WebSocketError(None, None),
        # could happen but Async PRAW should handle it
    )
    async def test_submit_image__timeout_4(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            image = image_path("test.jpg")
            with pytest.raises(WebSocketException):
                await subreddit.submit_image("Test Title", image)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_image__without_websockets(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for file_name in ("test.png", "test.jpg", "test.gif"):
                image = image_path(file_name)
                submission = await subreddit.submit_image(
                    "Test Title", image, without_websockets=True
                )
                assert submission is None

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        return_value=WebsocketMock("l6g58s", "l6g5al"),  # update with cassette
    )
    async def test_submit_video(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for i, file_name in enumerate(("test.mov", "test.mp4")):
                video = image_path(file_name)
                submission = await subreddit.submit_video(f"Test Title {i}", video)
                await submission.load()
                assert submission.author == pytest.placeholders.username
                # assert submission.is_reddit_media_domain
                # for some reason returns false
                assert submission.title == f"Test Title {i}"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_video__bad_filetype(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for file_name in ("test.jpg", "test.png", "test.gif"):
                video = image_path(file_name)
                with pytest.raises(ClientException):
                    await subreddit.submit_video("Test Title", video)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch("aiohttp.client.ClientSession.ws_connect", return_value=WebsocketMock())
    async def test_submit_video__bad_websocket(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette("TestSubreddit.test_submit_video"):
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for file_name in ("test.mov", "test.mp4"):
                video = image_path(file_name)
                with pytest.raises(ClientException):
                    await subreddit.submit_video("Test Title", video)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect", return_value=WebsocketMock("l6g771")
    )  # update with cassette
    async def test_submit_video__flair(self, _, __):
        flair_id = "6fc213da-cae7-11ea-9274-0e2407099e45"
        flair_text = "Test flair text"
        flair_class = ""
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            video = image_path("test.mov")
            submission = await subreddit.submit_video(
                "Test Title", video, flair_id=flair_id, flair_text=flair_text
            )
            assert submission.link_flair_css_class == flair_class
            assert submission.link_flair_text == flair_text

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect", return_value=WebsocketMock("l6gocy")
    )  # update with cassette
    async def test_submit_video_chat(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            video = image_path("test.mov")
            submission = await subreddit.submit_video(
                "Test Title", video, discussion_type="CHAT"
            )
            await submission.load()
            assert submission.discussion_type == "CHAT"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_video_verify_invalid(self, _):
        self.reddit.read_only = False
        self.reddit.validate_on_submit = True
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            video = image_path("test.mov")
            with pytest.raises(
                (RedditAPIException, BadRequest)
            ):  # waiting for prawcore fix
                await subreddit.submit_video(
                    "gdfgfdgdgdgfgfdgdfgfdgfdg", video, without_websockets=True
                )

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        return_value=WebsocketMock("l6gvvi", "l6gvx7"),  # update with cassette
    )
    async def test_submit_video__thumbnail(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for video_name, thumb_name in (
                ("test.mov", "test.jpg"),
                ("test.mp4", "test.png"),
            ):
                video = image_path(video_name)
                thumb = image_path(thumb_name)
                submission = await subreddit.submit_video(
                    "Test Title", video, thumbnail_path=thumb
                )
                await submission.load()
                assert submission.author == pytest.placeholders.username
                assert submission.is_video
                assert submission.title == "Test Title"

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect", side_effect=BlockingIOError
    )  # happens with timeout=0
    async def test_submit_video__timeout_1(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            video = image_path("test.mov")
            with pytest.raises(WebSocketException):
                await subreddit.submit_video("Test Title", video)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        side_effect=socket.timeout
        # happens with timeout=0.00001
    )
    async def test_submit_video__timeout_2(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            video = image_path("test.mov")
            with pytest.raises(WebSocketException):
                await subreddit.submit_video("Test Title", video)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        side_effect=TimeoutError,
        # could happen but Async PRAW should handle it
    )
    async def test_submit_video__timeout_3(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            video = image_path("test.mov")
            with pytest.raises(WebSocketException):
                await subreddit.submit_video("Test Title", video)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        side_effect=WebSocketError(None, None),
        # could happen but Async PRAW should handle it
    )
    async def test_submit_video__timeout_4(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            video = image_path("test.mov")
            with pytest.raises(WebSocketException):
                await subreddit.submit_video("Test Title", video)

    @mock.patch("asyncio.sleep", return_value=None)
    @mock.patch(
        "aiohttp.client.ClientSession.ws_connect",
        return_value=WebsocketMock("l6gtwa", "l6gty1"),  # update with cassette
    )
    async def test_submit_video__videogif(self, _, __):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for file_name in ("test.mov", "test.mp4"):
                video = image_path(file_name)
                submission = await subreddit.submit_video(
                    "Test Title", video, videogif=True
                )
                assert submission.author == pytest.placeholders.username
                assert submission.is_video
                assert submission.title == "Test Title"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit_video__without_websockets(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for file_name in ("test.mov", "test.mp4"):
                video = image_path(file_name)
                submission = await subreddit.submit_video(
                    "Test Title", video, without_websockets=True
                )
                assert submission is None

    async def test_subscribe(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.subscribe()

    async def test_subscribe__multiple(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.subscribe(
                other_subreddits=["redditdev", await self.reddit.subreddit("iama")]
            )

    async def test_traffic(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            traffic = await subreddit.traffic()
            assert isinstance(traffic, dict)

    async def test_traffic__not_public(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("announcements")
        with self.use_cassette():
            with pytest.raises(NotFound):
                await subreddit.traffic()

    async def test_unsubscribe(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.unsubscribe()

    async def test_unsubscribe__multiple(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.unsubscribe(
                other_subreddits=["redditdev", await self.reddit.subreddit("iama")]
            )


class TestSubredditFilters(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test__aiter__all(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("all")
            filters = await self.async_list(subreddit.filters)
        assert len(filters) > 0
        assert all(isinstance(x, Subreddit) for x in filters)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test__aiter__mod(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("mod")
        with self.use_cassette():
            filters = await self.async_list(subreddit.filters)
        assert len(filters) > 0
        assert all(isinstance(x, Subreddit) for x in filters)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("all")
            await subreddit.filters.add("redditdev")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add__subreddit_model(self, _):
        self.reddit.read_only = False
        with self.use_cassette("TestSubredditFilters.test_add"):
            subreddit = await self.reddit.subreddit("all")
            await subreddit.filters.add(await self.reddit.subreddit("redditdev"))

    # @mock.patch("asyncio.sleep", return_value=None) # FIXME: no longer raises not found; same with praw
    # async def test_add__non_special(self, _):
    #     self.reddit.read_only = False
    #     with self.use_cassette():
    #         with pytest.raises(NotFound):
    #             subreddit = await self.reddit.subreddit("redditdev")
    #             await subreddit.filters.add("redditdev")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_remove(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("mod")
            await subreddit.filters.remove("redditdev")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_remove__subreddit_model(self, _):
        self.reddit.read_only = False
        with self.use_cassette("TestSubredditFilters.test_remove"):
            subreddit = await self.reddit.subreddit("mod")
            await subreddit.filters.remove(await self.reddit.subreddit("redditdev"))

    # @mock.patch("asyncio.sleep", return_value=None) # FIXME: no longer rases not found; same with praw
    # async def test_remove__non_special(self, _):
    #     self.reddit.read_only = False
    #     with self.use_cassette():
    #         with pytest.raises(NotFound):
    #             subreddit = await self.reddit.subreddit("redditdev")
    #             await subreddit.filters.remove("redditdev")


class TestSubredditFlair(IntegrationTest):
    REDDITOR = pytest.placeholders.username

    async def test__call(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            mapping = subreddit.flair()
            assert len(await self.async_list(mapping)) > 0
            assert all([isinstance(x["user"], Redditor) async for x in mapping])

    async def test__call__user_filter(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            mapping = subreddit.flair(redditor=self.REDDITOR)
            assert len(await self.async_list(mapping)) == 1

    async def test_configure(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.configure(
                position=None,
                self_assign=True,
                link_position=None,
                link_self_assign=True,
            )

    async def test_configure__defaults(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.configure()

    async def test_delete(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.delete(pytest.placeholders.username)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete_all(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            response = await subreddit.flair.delete_all()
            assert len(response) == 5
            assert all("removed" in x["status"] for x in response)

    async def test_set__flair_id(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            flair = "c99ff6d0-c559-11ea-b93b-0ef0f80279f1"
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.set(
                redditor, flair_template_id=flair, text="redditor flair"
            )

    async def test_set__flair_id_default_text(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            flair = "c99ff6d0-c559-11ea-b93b-0ef0f80279f1"
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.set(redditor, flair_template_id=flair)

    async def test_set__redditor(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.set(redditor, text="redditor flair")

    async def test_set__redditor_css_only(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.set(
                pytest.placeholders.username, css_class="some class"
            )

    async def test_set__redditor_string(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.set(
                pytest.placeholders.username, css_class="some class", text="new flair"
            )

    async def test_update(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            flair_list = [
                redditor,
                "spez",
                {"user": "bsimpson"},
                {"user": "spladug", "flair_text": "", "flair_css_class": ""},
            ]
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            response = await subreddit.flair.update(
                flair_list, css_class="async default"
            )
        assert all(x["ok"] for x in response)
        assert not any(x["errors"] for x in response)
        assert not any(x["warnings"] for x in response)
        assert len([x for x in response if "added" in x["status"]]) == 3
        assert len([x for x in response if "removed" in x["status"]]) == 1
        for i, name in enumerate([str(redditor), "spez", "bsimpson", "spladug"]):
            assert name in response[i]["status"]

    async def test_update__comma_in_text(self):
        self.reddit.read_only = False
        flair_list = [
            {"user": "bsimpson"},
            {"user": "spladug", "flair_text": "a,b"},
        ]
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            response = await subreddit.flair.update(
                flair_list, css_class="async default"
            )
            assert all(x["ok"] for x in response)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_quotes(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            redditor = await self.reddit.user.me()
            response = await subreddit.flair.update(
                [redditor], css_class="testing", text='"testing"'
            )
            assert all(x["ok"] for x in response)
            flair = await self.async_next(subreddit.flair(redditor))
            assert flair["flair_text"] == '"testing"'
            assert flair["flair_css_class"] == "testing"


class TestSubredditFlairTemplates(IntegrationTest):
    async def test__aiter(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            templates = await self.async_list(subreddit.flair.templates)
        assert len(templates) == 1

        for flair_template in templates:
            assert flair_template["id"]

    async def test_add(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.templates.add(
                "PRAW", background_color="#ABCDEF", css_class="myCSS"
            )

    async def test_clear(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.templates.clear()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            template = next(iter(await self.async_list(subreddit.flair.templates)))
            await subreddit.flair.templates.delete(template["id"])

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            template = next(iter(await self.async_list(subreddit.flair.templates)))
            await subreddit.flair.templates.update(
                template["id"],
                background_color="#00FFFF",
                css_class="myCSS",
                text="PRAW updated",
                text_color="dark",
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_invalid(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(InvalidFlairTemplateID):
                subreddit = await self.reddit.subreddit(
                    pytest.placeholders.test_subreddit
                )
                await subreddit.flair.templates.update(
                    "fake id",
                    background_color="#00FFFF",
                    css_class="myCSS",
                    fetch=True,
                    text="PRAW updated",
                    text_color="dark",
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_fetch(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            template = next(iter(await self.async_list(subreddit.flair.templates)))
            await subreddit.flair.templates.update(
                template["id"],
                background_color="#00FFFF",
                css_class="myCSS",
                fetch=True,
                text="PRAW updated",
                text_color="dark",
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_fetch_no_css_class(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            template = next(iter(await self.async_list(subreddit.flair.templates)))
            await subreddit.flair.templates.update(
                template["id"],
                background_color="#00FFFF",
                fetch=True,
                text="PRAW updated",
                text_color="dark",
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_fetch_no_text(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            template = next(iter(await self.async_list(subreddit.flair.templates)))
            await subreddit.flair.templates.update(
                template["id"],
                background_color="#00FFFF",
                css_class="myCSS",
                fetch=True,
                text_color="dark",
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_fetch_no_text_or_css_class(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            template = next(iter(await self.async_list(subreddit.flair.templates)))
            await subreddit.flair.templates.update(
                template["id"],
                background_color="#00FFFF",
                fetch=True,
                text_color="dark",
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_fetch_only(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            template = next(iter(await self.async_list(subreddit.flair.templates)))
            await subreddit.flair.templates.update(template["id"], fetch=True)
            newtemplate = list(
                filter(
                    lambda _template: _template["id"] == template["id"],
                    [flair async for flair in subreddit.flair.templates],
                )
            )[0]
            assert newtemplate == template

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_false(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            template = next(iter(await self.async_list(subreddit.flair.templates)))
            await subreddit.flair.templates.update(
                template["id"], fetch=True, text_editable=True
            )
            await subreddit.flair.templates.update(
                template["id"], fetch=True, text_editable=False
            )
            newtemplate = list(
                filter(
                    lambda _template: _template["id"] == template["id"],
                    [flair async for flair in subreddit.flair.templates],
                )
            )[0]
            assert newtemplate["text_editable"] is False


class TestSubredditLinkFlairTemplates(IntegrationTest):
    async def test__aiter(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            templates = await self.async_list(subreddit.flair.link_templates)
        assert len(templates) == 2

        for template in templates:
            assert template["id"]
            assert isinstance(template["richtext"], list)
            assert all(isinstance(item, dict) for item in template["richtext"])

    async def test_add(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.link_templates.add(
                "PRAW", css_class="myCSS", text_color="light"
            )

    async def test_clear(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            await subreddit.flair.link_templates.clear()

    async def test_user_selectable(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            templates = await self.async_list(
                subreddit.flair.link_templates.user_selectable()
            )
        assert len(templates) >= 2

        for template in templates:
            assert template["flair_template_id"]
            assert template["flair_text"]


class TestSubredditListings(IntegrationTest):
    async def test_comments(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("askreddit")
            comments = await self.async_list(subreddit.comments())
        assert len(comments) == 100

    async def test_controversial(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("askreddit")
            submissions = await self.async_list(subreddit.controversial())
        assert len(submissions) == 100

    async def test_gilded(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("askreddit")
            submissions = await self.async_list(subreddit.gilded())
        assert len(submissions) >= 50

    async def test_hot(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("askreddit")
            submissions = await self.async_list(subreddit.hot())
        assert len(submissions) == 100

    async def test_new(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("askreddit")
            submissions = await self.async_list(subreddit.new())
        assert len(submissions) == 100

    async def test_random_rising(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("askreddit")
            submissions = await self.async_list(subreddit.random_rising())
        assert len(submissions) == 91

    async def test_rising(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("askreddit")
            submissions = await self.async_list(subreddit.rising())
        assert len(submissions) == 100

    async def test_top(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("askreddit")
            submissions = await self.async_list(subreddit.top())
        assert len(submissions) == 100


class TestSubredditModeration(IntegrationTest):
    async def test_accept_invite__no_invite(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                subreddit = await self.reddit.subreddit(
                    pytest.placeholders.test_subreddit
                )
                await subreddit.mod.accept_invite()
            assert excinfo.value.items[0].error_type == "NO_INVITE_FOUND"

    async def test_edited(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.edited():
                assert isinstance(item, (Comment, Submission))
                count += 1
            assert count == 100

    async def test_edited__only_comments(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.edited(only="comments"):
                assert isinstance(item, Comment)
                count += 1
            assert count == 100

    async def test_edited__only_submissions(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.edited(only="submissions"):
                assert isinstance(item, Submission)
                count += 1
            assert count > 0

    async def test_inbox(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit("all")
            async for item in subreddit.mod.inbox():
                assert isinstance(item, SubredditMessage)
                count += 1
            assert count == 100

    async def test_log(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit("mod")
            async for item in subreddit.mod.log():
                assert isinstance(item, ModAction)
                count += 1
            assert count == 100

    async def test_log__filters(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            redditor = await self.reddit.user.me()
            subreddit = await self.reddit.subreddit("mod")
            async for item in subreddit.mod.log(action="invitemoderator", mod=redditor):
                assert isinstance(item, ModAction)
                assert item.action == "invitemoderator"
                assert isinstance(item.mod, Redditor)
                assert item.mod == pytest.placeholders.username
                count += 1
            assert count > 0

    async def test_modqueue(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.modqueue():
                assert isinstance(item, (Comment, Submission))
                count += 1
            assert count > 0

    async def test_modqueue__only_comments(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.modqueue(only="comments"):
                assert isinstance(item, Comment)
                count += 1
            assert count > 0

    async def test_modqueue__only_submissions(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.modqueue(only="submissions"):
                assert isinstance(item, Submission)
                count += 1
            assert count > 0

    async def test_notes(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            notes = await self.async_list(
                subreddit.mod.notes.redditors("Watchful1", "watchful12", "spez")
            )
            assert len(notes) == 3
            assert notes[0].user.name.lower() == "watchful1"
            assert notes[1].user.name.lower() == "watchful12"
            assert notes[2] is None

    async def test_reports(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.reports():
                assert isinstance(item, (Comment, Submission))
                count += 1
            assert count == 100

    async def test_reports__only_comments(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.reports(only="comments"):
                assert isinstance(item, Comment)
                count += 1
            assert count > 0

    async def test_reports__only_submissions(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.reports(only="submissions"):
                assert isinstance(item, Submission)
                count += 1
            assert count == 100

    async def test_spam(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.spam():
                assert isinstance(item, (Comment, Submission))
                count += 1
            assert count > 0

    async def test_spam__only_comments(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.spam(only="comments"):
                assert isinstance(item, Comment)
                count += 1
            assert count > 0

    async def test_spam__only_submissions(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.spam(only="submissions"):
                assert isinstance(item, Submission)
                count += 1
            assert count > 0

    async def test_unmoderated(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for item in subreddit.mod.unmoderated():
                assert isinstance(item, (Comment, Submission))
                count += 1
            assert count > 0

    async def test_unread(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit("all")
            async for item in subreddit.mod.unread():
                assert isinstance(item, SubredditMessage)
                count += 1
            assert count > 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            before_settings = await subreddit.mod.settings()
            new_title = f"{before_settings['title']}x"
            new_title = (
                "x"
                if (len(new_title) >= 20 and "placeholder" not in new_title)
                else new_title
            )
            await subreddit.mod.update(title=new_title)
            await subreddit.load()
            assert subreddit.title == new_title
            after_settings = await subreddit.mod.settings()

            # Ensure that nothing has changed besides what was specified.
            before_settings["title"] = new_title
            assert before_settings == after_settings


class TestSubredditModmail(IntegrationTest):
    async def test_bulk_read(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            conversations = await subreddit.modmail.bulk_read(state="new")
            for conversation in conversations:
                assert isinstance(conversation, ModmailConversation)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_call(self, _):
        self.reddit.read_only = False
        conversation_id = "fjhla"
        subreddit = await self.reddit.subreddit("all")
        with self.use_cassette():
            conversation = await subreddit.modmail(conversation_id)
            assert isinstance(conversation.user, Redditor)
            for message in conversation.messages:
                assert isinstance(message, ModmailMessage)
            for action in conversation.mod_actions:
                assert isinstance(action, ModmailAction)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_call__internal(self, _):
        self.reddit.read_only = False
        conversation_id = "ff1r8"
        subreddit = await self.reddit.subreddit("all")
        with self.use_cassette():
            conversation = await subreddit.modmail(conversation_id)
            for message in conversation.messages:
                assert isinstance(message, ModmailMessage)
            for action in conversation.mod_actions:
                assert isinstance(action, ModmailAction)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_call__mark_read(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("all")
        with self.use_cassette():
            conversation = await subreddit.modmail("fccdg", mark_read=True)
            assert conversation.last_unread is None

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_conversations(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("all")
            conversations = subreddit.modmail.conversations()
            async for conversation in conversations:
                assert isinstance(conversation, ModmailConversation)
                assert isinstance(conversation.authors[0], Redditor)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_conversations__other_subreddits(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("pics")
            conversations = subreddit.modmail.conversations(
                other_subreddits=["LifeProTips"]
            )
            assert (
                len(set([conversation.owner async for conversation in conversations]))
                > 1
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_conversations__params(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("all")
            conversations = subreddit.modmail.conversations(state="mod")
            async for conversation in conversations:
                assert conversation.is_internal

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            conversation = await subreddit.modmail.create(
                subject="Subject",
                body="Body",
                recipient=redditor,
            )
        assert isinstance(conversation, ModmailConversation)

    async def test_subreddits(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            async for subreddit in subreddit.modmail.subreddits():
                assert isinstance(subreddit, Subreddit)

    async def test_unread_count(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            assert isinstance(await subreddit.modmail.unread_count(), dict)


class TestSubredditQuarantine(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_opt_in(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("tiananmenaquarefalse")
        with self.use_cassette():
            with pytest.raises(Forbidden):
                await self.async_next(subreddit.top())
            await subreddit.quaran.opt_in()
            assert isinstance(await self.async_next(subreddit.top()), Submission)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_opt_out(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("tiananmenaquarefalse")
        with self.use_cassette():
            await subreddit.quaran.opt_out()
            with pytest.raises(Forbidden):
                await self.async_next(subreddit.new())


class TestSubredditRelationships(IntegrationTest):
    REDDITOR = "pyapitestuser3"

    async def add_remove(self, base, user, relationship):
        relationship = getattr(base, relationship)
        await relationship.add(user)
        relationships = await self.async_list(relationship())
        assert user in relationships
        redditor = relationships[relationships.index(user)]
        assert isinstance(redditor, Redditor)
        assert hasattr(redditor, "date")
        await relationship.remove(user)
        assert user not in await self.async_list(relationship())

    async def test_banned(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await self.add_remove(subreddit, self.REDDITOR, "banned")

    async def test_banned__user_filter(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        banned = subreddit.banned(redditor="pyapitestuser3")
        with self.use_cassette():
            assert len(await self.async_list(banned)) == 1

    async def test_contributor(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await self.add_remove(subreddit, self.REDDITOR, "contributor")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_contributor_leave(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.contributor.leave()

    async def test_contributor__user_filter(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        contributor = subreddit.contributor(redditor="pyapitestuser3")
        with self.use_cassette():
            assert len(await self.async_list(contributor)) == 1

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_moderator(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            # Moderators can only be invited.
            # As of 2016-03-18 there is no API endpoint to get the moderator
            # invite list.
            await subreddit.moderator.add(self.REDDITOR)
            assert self.REDDITOR not in await subreddit.moderator()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_moderator_aiter(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            async for moderator in subreddit.moderator:
                assert isinstance(moderator, Redditor)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_moderator__limited_permissions(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            # Moderators can only be invited.
            # As of 2016-03-18 there is no API endpoint to get the moderator
            # invite list.
            await subreddit.moderator.add(self.REDDITOR, permissions=["access", "wiki"])
            assert self.REDDITOR not in await subreddit.moderator()

    async def test_moderator_invite__invalid_perm(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await subreddit.moderator.invite(self.REDDITOR, permissions=["a"])
            assert excinfo.value.items[0].error_type == "INVALID_PERMISSIONS"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_moderator_invite__no_perms(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            # Moderators can only be invited.
            # As of 2016-03-18 there is no API endpoint to get the moderator
            # invite list.
            await subreddit.moderator.invite(self.REDDITOR, permissions=[])
            assert self.REDDITOR not in await subreddit.moderator()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_moderator_invited_moderators(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            invited = subreddit.moderator.invited()
            assert isinstance(invited, ListingGenerator)
            async for moderator in invited:
                assert isinstance(moderator, Redditor)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_moderator_leave(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.moderator.leave()

    async def test_moderator_update(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.moderator.update(
                pytest.placeholders.username, permissions=["config"]
            )

    async def test_moderator_update_invite(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.moderator.update_invite(self.REDDITOR, permissions=["mail"])

    async def test_moderator__user_filter(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            moderator = await subreddit.moderator(redditor=pytest.placeholders.username)
        assert len(moderator) == 1
        assert "mod_permissions" in moderator[0].__dict__

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_muted(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await self.add_remove(subreddit, self.REDDITOR, "muted")

    async def test_moderator_remove_invite(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.moderator.remove_invite(self.REDDITOR)

    async def test_wiki_banned(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await self.add_remove(subreddit.wiki, self.REDDITOR, "banned")

    async def test_wiki_contributor(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await self.add_remove(subreddit.wiki, self.REDDITOR, "contributor")


class TestSubredditStreams(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_comments(self, _):
        subreddit = await self.reddit.subreddit("all")
        with self.use_cassette():
            generator = subreddit.stream.comments()
            for i in range(400):
                assert isinstance(await self.async_next(generator), Comment)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_comments__with_pause(self, _):
        subreddit = await self.reddit.subreddit("askreddit")
        with self.use_cassette():
            comment_stream = subreddit.stream.comments(pause_after=0)
            comment_count = 1
            pause_count = 1
            comment = await self.async_next(comment_stream)
            while comment is not None:
                comment_count += 1
                comment = await self.async_next(comment_stream)
            while comment is None:
                pause_count += 1
                comment = await self.async_next(comment_stream)
            assert comment_count == 108
            assert pause_count == 3

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_comments__with_skip_existing(self, _):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("askreddit")
            generator = subreddit.stream.comments(skip_existing=True, pause_after=-1)
            comment = await self.async_next(generator)
            assert comment is None
            comment = await self.async_next(generator)
            assert isinstance(comment, Comment)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submissions(self, _):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("all")
            generator = subreddit.stream.submissions()
            for i in range(101):
                assert isinstance(await self.async_next(generator), Submission)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submissions__with_pause(self, _):
        with self.use_cassette("TestSubredditStreams.test_submissions"):
            subreddit = await self.reddit.subreddit("all")
            generator = subreddit.stream.submissions(pause_after=-1)
            submission = await self.async_next(generator)
            submission_count = 0
            while submission is not None:
                submission_count += 1
                submission = await self.async_next(generator)
            assert submission_count == 100

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submissions__with_pause_and_skip_after(self, _):
        with self.use_cassette("TestSubredditStreams.test_submissions"):
            subreddit = await self.reddit.subreddit("all")
            generator = subreddit.stream.submissions(pause_after=-1, skip_existing=True)
            submission = await self.async_next(generator)
            assert submission is None  # The first thing yielded should be None
            submission_count = 0
            submission = await self.async_next(generator)
            while submission is not None:
                submission_count += 1
                submission = await self.async_next(generator)
            assert submission_count < 100


class TestSubredditModerationStreams(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_edited(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("mod")
        with self.use_cassette():
            generator = subreddit.mod.stream.edited()
            for i in range(101):
                assert isinstance(
                    await self.async_next(generator), (Comment, Submission)
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_log(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("mod")
        with self.use_cassette():
            generator = subreddit.mod.stream.log()
            for i in range(101):
                assert isinstance(await self.async_next(generator), ModAction)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_modmail_conversations(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("mod")
        with self.use_cassette():
            generator = subreddit.mod.stream.modmail_conversations()
            for i in range(101):
                assert isinstance(await self.async_next(generator), ModmailConversation)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_modqueue(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("mod")
        with self.use_cassette():
            generator = subreddit.mod.stream.modqueue()
            for i in range(10):
                assert isinstance(
                    await self.async_next(generator), (Comment, Submission)
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_spam(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("mod")
        with self.use_cassette():
            generator = subreddit.mod.stream.spam()
            for i in range(101):
                assert isinstance(
                    await self.async_next(generator), (Comment, Submission)
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_reports(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("mod")
        with self.use_cassette():
            generator = subreddit.mod.stream.reports()
            for i in range(10):
                assert isinstance(
                    await self.async_next(generator), (Comment, Submission)
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unmoderated(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("mod")
        with self.use_cassette():
            generator = subreddit.mod.stream.unmoderated()
            for i in range(101):
                assert isinstance(
                    await self.async_next(generator), (Comment, Submission)
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unread(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit("mod")
        with self.use_cassette():
            generator = subreddit.mod.stream.unread()
            for i in range(2):
                assert isinstance(await self.async_next(generator), SubredditMessage)


class TestSubredditStylesheet(IntegrationTest):
    async def test_call(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            stylesheet = await subreddit.stylesheet()
        assert isinstance(stylesheet, Stylesheet)
        assert len(stylesheet.images) > 0
        assert stylesheet.stylesheet != ""

    async def test_delete_banner(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.delete_banner()

    async def test_delete_banner_additional_image(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.delete_banner_additional_image()

    async def test_delete_banner_hover_image(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.delete_banner_hover_image()

    async def test_delete_header(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.delete_header()

    async def test_delete_image(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.delete_image("praw")

    async def test_delete_mobile_header(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.delete_mobile_header()

    async def test_delete_mobile_icon(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.delete_mobile_icon()

    async def test_update(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.update("p { color: red; }")

    async def test_update__with_reason(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.update("div { color: red; }", reason="use div")

    async def test_upload(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            response = await subreddit.stylesheet.upload(
                name="asyncpraw", image_path=image_path("white-square.png")
            )
        assert response["img_src"].endswith(".png")

    async def test_upload__invalid(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await subreddit.stylesheet.upload(
                    name="asyncpraw", image_path=image_path("invalid.jpg")
                )
        assert excinfo.value.items[0].error_type == "IMAGE_ERROR"

    async def test_upload__invalid_ext(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await subreddit.stylesheet.upload(
                    name="asyncpraw", image_path=image_path("white-square.png")
                )
        assert excinfo.value.items[0].error_type == "BAD_CSS_NAME"

    async def test_upload__too_large(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            with pytest.raises(TooLarge):
                await subreddit.stylesheet.upload(
                    name="asyncpraw", image_path=image_path("too_large.jpg")
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_upload_banner__jpg(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.upload_banner(image_path("white-square.jpg"))

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_upload_banner__png(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.upload_banner(image_path("white-square.png"))

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_upload_banner_additional_image__jpg(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.upload_banner_additional_image(
                image_path("white-square.jpg")
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_upload_banner_additional_image__png(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.upload_banner_additional_image(
                image_path("white-square.png")
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_upload_banner_additional_image__align(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            for alignment in ("left", "centered", "right"):
                await subreddit.stylesheet.upload_banner_additional_image(
                    image_path("white-square.png"), align=alignment
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_upload_banner_hover_image__jpg(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.upload_banner_additional_image(
                image_path("white-square.png")
            )
            await subreddit.stylesheet.upload_banner_hover_image(
                image_path("white-square.jpg")
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_upload_banner_hover_image__png(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            await subreddit.stylesheet.upload_banner_additional_image(
                image_path("white-square.jpg")
            )
            await subreddit.stylesheet.upload_banner_hover_image(
                image_path("white-square.png")
            )

    async def test_upload_header__jpg(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            response = await subreddit.stylesheet.upload_header(
                image_path("white-square.jpg")
            )
        assert response["img_src"].endswith(".jpg")

    async def test_upload_header__png(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            response = await subreddit.stylesheet.upload_header(
                image_path("white-square.png")
            )
        assert response["img_src"].endswith(".png")

    async def test_upload_mobile_header(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            response = await subreddit.stylesheet.upload_mobile_header(
                image_path("header.jpg")
            )
        assert response["img_src"].endswith(".jpg")

    async def test_upload_mobile_icon(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            response = await subreddit.stylesheet.upload_mobile_icon(
                image_path("icon.jpg")
            )
        assert response["img_src"].endswith(".jpg")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_upload__others_invalid(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            for method in [
                "upload_header",
                "upload_mobile_header",
                "upload_mobile_icon",
            ]:
                with pytest.raises(RedditAPIException) as excinfo:
                    await getattr(subreddit.stylesheet, method)(
                        image_path("invalid.jpg")
                    )
                assert excinfo.value.items[0].error_type == "IMAGE_ERROR"

    async def test_upload__others_too_large(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            for method in [
                "upload_header",
                "upload_mobile_header",
                "upload_mobile_icon",
            ]:
                with pytest.raises(TooLarge):
                    await getattr(
                        subreddit.stylesheet,
                        method,
                    )(image_path("too_large.jpg"))


class TestSubredditWiki(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test__aiter(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            count = 0
            async for wikipage in subreddit.wiki:
                assert isinstance(wikipage, WikiPage)
                count += 1
            assert count > 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        with self.use_cassette():
            wikipage = await subreddit.wiki.create(
                name="Async PRAW New Page", content="This is the new wiki page"
            )
            await wikipage.load()
            assert wikipage.name == "async_praw_new_page"
            assert wikipage.content_md == "This is the new wiki page"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_revisions(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        with self.use_cassette():
            count = 0
            async for revision in subreddit.wiki.revisions(limit=2):
                count += 1
                assert isinstance(revision["author"], Redditor)
                assert isinstance(revision["page"], WikiPage)
            assert count == 2



import pytest

from asyncpraw import Reddit
from asyncpraw.exceptions import ReadOnlyException
from asyncpraw.models import User

from .. import UnitTest


class TestUser(UnitTest):
    async def test_me__in_read_only_mode(self):
        async with Reddit(
            client_id="dummy",
            client_secret="dummy",
            praw8_raise_exception_on_me=True,
            user_agent="dummy",
        ) as reddit:
            assert reddit.read_only
            user = User(reddit)
            with pytest.raises(ReadOnlyException):
                await user.me()

    async def test_me__in_read_only_mode__deprecated(self, reddit):
        assert reddit.read_only
        assert await User(reddit).me() is None



"""Provide the MessageableMixin class."""
from typing import TYPE_CHECKING, Optional, Union

from ....const import API_PATH
from ....util import _deprecate_args

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class MessageableMixin:
    """Interface for classes that can be messaged."""

    @_deprecate_args("subject", "message", "from_subreddit")
    async def message(
        self,
        *,
        from_subreddit: Optional[Union["asyncpraw.models.Subreddit", str]] = None,
        message: str,
        subject: str,
    ):
        """Send a message to a :class:`.Redditor` or a :class:`.Subreddit`'s moderators (modmail).

        :param from_subreddit: A :class:`.Subreddit` instance or string to send the
            message from. When provided, messages are sent from the subreddit rather
            than from the authenticated user.

            .. note::

                The authenticated user must be a moderator of the subreddit and have the
                ``mail`` moderator permission.

        :param message: The message content.
        :param subject: The subject of the message.

        For example, to send a private message to u/spez, try:

        .. code-block:: python

            redditor = await reddit.redditor("spez", fetch=False)
            await redditor.message(subject="TEST", message="test message from Async PRAW")

        To send a message to u/spez from the moderators of r/test try:

        .. code-block:: python

            redditor = await reddit.redditor("spez", fetch=False)
            await redditor.message(
                subject="TEST", message="test message from r/test", from_subreddit="test"
            )

        To send a message to the moderators of r/test, try:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            await subreddit.message(subject="TEST", message="test PM from Async PRAW")

        """
        data = {
            "subject": subject,
            "text": message,
            "to": f"{getattr(self.__class__, 'MESSAGE_PREFIX', '')}{self}",
        }
        if from_subreddit:
            data["from_sr"] = str(from_subreddit)
        await self._reddit.post(API_PATH["compose"], data=data)



"""Provide the helper classes."""
from json import dumps
from typing import TYPE_CHECKING, AsyncGenerator, List, Optional, Union

from ..const import API_PATH
from ..util import _deprecate_args
from .base import AsyncPRAWBase
from .reddit.draft import Draft
from .reddit.live import LiveThread
from .reddit.multi import Multireddit, Subreddit

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class DraftHelper(AsyncPRAWBase):
    r"""Provide a set of functions to interact with :class:`.Draft` instances.

    .. note::

        The methods provided by this class will only work on the currently authenticated
        user's :class:`.Draft`\ s.

    """

    async def __call__(
        self, draft_id: Optional[str] = None, fetch: bool = True
    ) -> Union[List["asyncpraw.models.Draft"], "asyncpraw.models.Draft"]:
        """Return a list of :class:`.Draft` instances.

        :param draft_id: When provided, this returns a :class:`.Draft` instance
            (default: ``None``).
        :param fetch: Determines if Async PRAW will fetch the object (default:
            ``True``).

        :returns: A :class:`.Draft` instance if ``draft_id`` is provided. Otherwise, a
            list of :class:`.Draft` objects.

        This method can be used to fetch a specific draft by ID, like so:

        .. code-block:: python

            draft_id = "124862bc-e1e9-11eb-aa4f-e68667a77cbb"
            draft = await reddit.drafts(draft_id)
            print(draft)

        """
        if draft_id is not None:
            draft = Draft(self._reddit, id=draft_id)
            if fetch:
                await draft.load()
            return draft
        return await self._draft_list()

    async def __aiter__(self):
        r"""Iterate through all the :class:`.Draft`\ s.

        :returns: An asynchronous iterator containing all the currently authenticated
            user's :class:`.Draft`\ s.

        .. code-block:: python

            async for draft in reddit.drafts:
                print(draft)

        """
        drafts = await self._draft_list()
        for draft in drafts:
            yield draft

    async def _draft_list(self) -> List["asyncpraw.models.Draft"]:
        """Get a list of :class:`.Draft` instances.

        :returns: A list of :class:`.Draft` instances.

        """
        return await self._reddit.get(API_PATH["drafts"], params={"md_body": True})

    async def create(
        self,
        *,
        flair_id: Optional[str] = None,
        flair_text: Optional[str] = None,
        is_public_link: bool = False,
        nsfw: bool = False,
        original_content: bool = False,
        selftext: Optional[str] = None,
        send_replies: bool = True,
        spoiler: bool = False,
        subreddit: Optional[
            Union[str, "asyncpraw.models.Subreddit", "asyncpraw.models.UserSubreddit"]
        ] = None,
        title: Optional[str] = None,
        url: Optional[str] = None,
        **draft_kwargs,
    ) -> "asyncpraw.models.Draft":
        """Create a new :class:`.Draft`.

        :param flair_id: The flair template to select (default: ``None``).
        :param flair_text: If the template's ``flair_text_editable`` value is ``True``,
            this value will set a custom text (default: ``None``). ``flair_id`` is
            required when ``flair_text`` is provided.
        :param is_public_link: Whether to enable public viewing of the draft before it
            is submitted (default: ``False``).
        :param nsfw: Whether the draft should be marked NSFW (default: ``False``).
        :param original_content: Whether the submission should be marked as original
            content (default: ``False``).
        :param selftext: The Markdown formatted content for a text submission draft. Use
            ``None`` to make a title-only submission draft (default: ``None``).
            ``selftext`` can not be provided if ``url`` is provided.
        :param send_replies: When ``True``, messages will be sent to the submission
            author when comments are made to the submission (default: ``True``).
        :param spoiler: Whether the submission should be marked as a spoiler (default:
            ``False``).
        :param subreddit: The subreddit to create the draft for. This accepts a
            subreddit display name, :class:`.Subreddit` object, or
            :class:`.UserSubreddit` object. If ``None``, the :class:`.UserSubreddit` of
            currently authenticated user will be used (default: ``None``).
        :param title: The title of the draft (default: ``None``).
        :param url: The URL for a ``link`` submission draft (default: ``None``). ``url``
            can not be provided if ``selftext`` is provided.

        Additional keyword arguments can be provided to handle new parameters as Reddit
        introduces them.

        :returns: The new :class:`.Draft` object.

        """
        if selftext and url:
            raise TypeError("Exactly one of 'selftext' or 'url' must be provided.")
        if isinstance(subreddit, str):
            subreddit = await self._reddit.subreddit(subreddit)

        data = await Draft._prepare_data(
            flair_id=flair_id,
            flair_text=flair_text,
            is_public_link=is_public_link,
            nsfw=nsfw,
            original_content=original_content,
            selftext=selftext,
            send_replies=send_replies,
            spoiler=spoiler,
            subreddit=subreddit,
            title=title,
            url=url,
            **draft_kwargs,
        )
        return await self._reddit.post(API_PATH["draft"], data=data)


class LiveHelper(AsyncPRAWBase):
    r"""Provide a set of functions to interact with :class:`.LiveThread`\ s."""

    async def __call__(
        self, id: str, fetch: bool = False
    ) -> "asyncpraw.models.LiveThread":  # pylint: disable=invalid-name,redefined-builtin
        """Return a new instance of :class:`.LiveThread`.

        This method is intended to be used as:

        .. code-block:: python

            livethread = await reddit.live("ukaeu1ik4sw5")

        If you need the object fetched right away (e.g., to access attributes) you can
        do:

        .. code-block:: python

            livethread = await reddit.live("ukaeu1ik4sw5", fetch=True)
            await livethread.close()

        :param id: A live thread ID, e.g., ``ukaeu1ik4sw5``.
        :param fetch: Determines if Async PRAW will fetch the object (default:
            ``False``).

        """
        live_thread = LiveThread(self._reddit, id=id)
        if fetch:
            await live_thread._fetch()
        return live_thread

    def info(
        self, ids: List[str]
    ) -> AsyncGenerator["asyncpraw.models.LiveThread", None]:
        """Fetch information about each live thread in ``ids``.

        :param ids: A list of IDs for a live thread.

        :returns: A generator that yields :class:`.LiveThread` instances.

        :raises: ``asyncprawcore.ServerError`` if invalid live threads are requested.

        Requests will be issued in batches for each 100 IDs.

        .. note::

            This method doesn't support IDs for live updates.

        .. warning::

            Unlike :meth:`.Reddit.info`, the output of this method may not reflect the
            order of input.

        Usage:

        .. code-block:: python

            ids = ["3rgnbke2rai6hen7ciytwcxadi", "sw7bubeycai6hey4ciytwamw3a", "t8jnufucss07"]
            async for thread in reddit.live.info(ids):
                print(thread.title)

        """
        if not isinstance(ids, list):
            raise TypeError("ids must be a list")

        async def generator():
            for position in range(0, len(ids), 100):
                ids_chunk = ids[position : position + 100]
                url = API_PATH["live_info"].format(ids=",".join(ids_chunk))
                params = {"limit": 100}  # 25 is used if not specified
                for result in await self._reddit.get(url, params=params):
                    yield result

        return generator()

    @_deprecate_args("title", "description", "nsfw", "resources")
    async def create(
        self,
        title: str,
        *,
        description: Optional[str] = None,
        nsfw: bool = False,
        resources: str = None,
    ) -> "asyncpraw.models.LiveThread":
        """Create a new :class:`.LiveThread`.

        :param title: The title of the new :class:`.LiveThread`.
        :param description: The new :class:`.LiveThread`'s description.
        :param nsfw: Indicate whether this thread is not safe for work (default:
            ``False``).
        :param resources: Markdown formatted information that is useful for the
            :class:`.LiveThread`.

        :returns: The new :class:`.LiveThread` object.

        """
        return await self._reddit.post(
            API_PATH["livecreate"],
            data={
                "description": description,
                "nsfw": nsfw,
                "resources": resources,
                "title": title,
            },
        )

    async def now(self) -> Optional["asyncpraw.models.LiveThread"]:
        """Get the currently featured live thread.

        :returns: The :class:`.LiveThread` object, or ``None`` if there is no currently
            featured live thread.

        Usage:

        .. code-block:: python

            thread = await reddit.live.now()  # LiveThread object or None

        """
        return await self._reddit.get(API_PATH["live_now"])


class MultiredditHelper(AsyncPRAWBase):
    """Provide a set of functions to interact with multireddits."""

    @_deprecate_args("redditor", "name")
    async def __call__(
        self,
        *,
        name: str,
        redditor: Union[str, "asyncpraw.models.Redditor"],
        fetch: bool = False,
    ) -> "asyncpraw.models.Multireddit":
        """Return a lazy instance of :class:`.Multireddit`.

        If you need the object fetched right away (e.g., to access an attribute) you can
        do:

        .. code-block:: python

            multireddit = await reddit.multireddit(redditor="redditor", name="multi", fetch=True)
            async for comment in multireddit.comments(limit=25):
                print(comment.author)

        :param name: The name of the multireddit.
        :param redditor: A redditor name or :class:`.Redditor` instance who owns the
            multireddit.
        :param name: The name of the multireddit.
        :param fetch: Determines if Async PRAW will fetch the object (default:
            ``False``).

        """
        path = f"/user/{redditor}/m/{name}"
        multireddit = Multireddit(self._reddit, _data={"name": name, "path": path})
        if fetch:
            await multireddit._fetch()
        return multireddit

    @_deprecate_args(
        "display_name",
        "subreddits",
        "description_md",
        "icon_name",
        "key_color",
        "visibility",
        "weighting_scheme",
    )
    async def create(
        self,
        *,
        description_md: Optional[str] = None,
        display_name: str,
        icon_name: Optional[str] = None,
        key_color: Optional[str] = None,
        subreddits: Union[str, "asyncpraw.models.Subreddit"],
        visibility: str = "private",
        weighting_scheme: str = "classic",
    ) -> "asyncpraw.models.Multireddit":
        """Create a new :class:`.Multireddit`.

        :param display_name: The display name for the new multireddit.
        :param subreddits: Subreddits to add to the new multireddit. Can be a list of
            either :class:`.Subreddit` instances or subreddit display names.
        :param description_md: Description for the new multireddit, formatted in
            markdown.
        :param icon_name: Can be one of: ``"art and design"``, ``"ask"``, ``"books"``,
            ``"business"``, ``"cars"``, ``"comics"``, ``"cute animals"``, ``"diy"``,
            ``"entertainment"``, ``"food and drink"``, ``"funny"``, ``"games"``,
            ``"grooming"``, ``"health"``, ``"life advice"``, ``"military"``, ``"models
            pinup"``, ``"music"``, ``"news"``, ``"philosophy"``, ``"pictures and
            gifs"``, ``"science"``, ``"shopping"``, ``"sports"``, ``"style"``,
            ``"tech"``, ``"travel"``, ``"unusual stories"``, ``"video"``, or ``None``.
        :param key_color: RGB hex color code of the form ``"#FFFFFF"``.
        :param visibility: Can be one of: ``"hidden"``, ``"private"``, or ``"public"``
            (default: ``"private"``).
        :param weighting_scheme: Can be one of: ``"classic"`` or ``"fresh"`` (default:
            ``"classic"``).

        :returns: The new :class:`.Multireddit` object.

        """
        model = {
            "description_md": description_md,
            "display_name": display_name,
            "icon_name": icon_name,
            "key_color": key_color,
            "subreddits": [{"name": str(sub)} for sub in subreddits],
            "visibility": visibility,
            "weighting_scheme": weighting_scheme,
        }
        return await self._reddit.post(
            API_PATH["multireddit_base"], data={"model": dumps(model)}
        )


class SubredditHelper(AsyncPRAWBase):
    """Provide a set of functions to interact with Subreddits."""

    async def __call__(
        self, display_name: str, fetch: bool = False
    ) -> "asyncpraw.models.Subreddit":
        """Return an instance of :class:`.Subreddit`.

        :param display_name: The name of the subreddit.
        :param fetch: Determines if Async PRAW will fetch the object (default:
            ``False``).

        If you need the object fetched right away (e.g., to access an attribute) you can
        do:

        .. code-block:: python

            subreddit = await reddit.subreddit("test", fetch=True)
            print(subreddit.subscribers)

        """
        lower_name = display_name.lower()

        if lower_name == "random":
            return await self._reddit.random_subreddit()
        if lower_name == "randnsfw":
            return await self._reddit.random_subreddit(nsfw=True)
        subreddit = Subreddit(self._reddit, display_name=display_name)
        if fetch:
            await subreddit._fetch()
        return subreddit

    @_deprecate_args("name", "title", "link_type", "subreddit_type", "wikimode")
    async def create(
        self,
        name: str,
        *,
        link_type: str = "any",
        subreddit_type: str = "public",
        title: Optional[str] = None,
        wikimode: str = "disabled",
        **other_settings: Optional[str],
    ) -> "asyncpraw.models.Subreddit":
        """Create a new :class:`.Subreddit`.

        :param name: The name for the new subreddit.
        :param link_type: The types of submissions users can make. One of ``"any"``,
            ``"link"``, or ``"self"`` (default: ``"any"``).
        :param subreddit_type: One of ``"archived"``, ``"employees_only"``,
            ``"gold_only"``, ``"gold_restricted"``, ``"private"``, ``"public"``, or
            ``"restricted"`` (default: ``"public"``).
        :param title: The title of the subreddit. When ``None`` or ``""`` use the value
            of ``name``.
        :param wikimode: One of ``"anyone"``, ``"disabled"``, or ``"modonly"`` (default:
            ``"disabled"``).

        Any keyword parameters not provided, or set explicitly to ``None``, will take on
        a default value assigned by the Reddit server.

        .. seealso::

            :meth:`~.SubredditModeration.update` for documentation of other available
            settings.

        """
        await Subreddit._create_or_update(
            _reddit=self._reddit,
            link_type=link_type,
            name=name,
            subreddit_type=subreddit_type,
            title=title or name,
            wikimode=wikimode,
            **other_settings,
        )
        subreddit = await self(name, fetch=True)
        return subreddit



import pytest

from asyncpraw.models import LiveThread, LiveUpdate, Redditor
from asyncpraw.models.reddit.live import (
    LiveContributorRelationship,
    LiveThreadContribution,
    LiveUpdateContribution,
)

from ... import UnitTest


class TestLiveThread(UnitTest):
    def test_construct_success(self, reddit):
        thread_id = "ukaeu1ik4sw5"
        data = {"id": thread_id}

        thread = LiveThread(reddit, thread_id)
        assert isinstance(thread, LiveThread)
        assert thread.id == thread_id

        thread = LiveThread(reddit, _data=data)
        assert isinstance(thread, LiveThread)
        assert thread.id == thread_id

    def test_construct_failure(self, reddit):
        message = "Either 'id' or '_data' must be provided."
        with pytest.raises(TypeError) as excinfo:
            LiveThread(reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            LiveThread(reddit, id="dummy", _data={"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(ValueError):
            LiveThread(reddit, "")

    def test_contrib(self, reddit):
        thread_id = "ukaeu1ik4sw5"
        thread = LiveThread(reddit, thread_id)
        assert isinstance(thread.contrib, LiveThreadContribution)

    def test_contributor(self, reddit):
        thread_id = "ukaeu1ik4sw5"
        thread = LiveThread(reddit, thread_id)
        assert isinstance(thread.contributor, LiveContributorRelationship)

    def test_equality(self, reddit):
        thread1 = LiveThread(reddit, id="dummy1")
        thread2 = LiveThread(reddit, id="Dummy1")
        thread3 = LiveThread(reddit, id="dummy3")
        assert thread1 == thread1
        assert thread2 == thread2
        assert thread3 == thread3
        assert thread1 != thread2  # live thread ID in a URL is case sensitive
        assert thread2 != thread3
        assert thread1 != thread3
        assert "dummy1" == thread1
        assert thread2 != "dummy1"
        assert thread2 == "Dummy1"

    async def test_getitem(self, reddit):
        thread_id = "dummy_thread_id"
        update_id = "dummy_update_id"
        thread = LiveThread(reddit, id=thread_id)
        update = await thread.get_update(update_id, fetch=False)
        assert isinstance(update, LiveUpdate)
        assert update.id == update_id

    def test_hash(self, reddit):
        thread1 = LiveThread(reddit, id="dummy1")
        thread2 = LiveThread(reddit, id="Dummy1")
        thread3 = LiveThread(reddit, id="dummy3")
        assert hash(thread1) == hash(thread1)
        assert hash(thread2) == hash(thread2)
        assert hash(thread3) == hash(thread3)
        assert hash(thread1) != hash(thread2)
        assert hash(thread2) != hash(thread3)
        assert hash(thread1) != hash(thread3)

    def test_repr(self, reddit):
        thread = LiveThread(reddit, id="dummy")
        assert repr(thread) == "LiveThread(id='dummy')"

    def test_str(self, reddit):
        thread = LiveThread(reddit, id="dummy")
        assert str(thread) == "dummy"


class TestLiveThreadContribution(UnitTest):
    async def test_update__no_args(self, reddit):
        thread = LiveThread(reddit, "xyu8kmjvfrww")
        assert await thread.contrib.update() is None


class TestLiveUpdate(UnitTest):
    def test_construct_success(self, reddit):
        thread_id = "dummy_thread_id"
        update_id = "dummy_update_id"
        data = {"id": update_id}

        update = LiveUpdate(reddit, thread_id=thread_id, update_id=update_id)
        assert isinstance(update, LiveUpdate)
        assert update.id == update_id
        assert isinstance(update.thread, LiveThread)
        assert update.thread.id == thread_id

        update = LiveUpdate(reddit, thread_id, update_id)
        assert isinstance(update, LiveUpdate)
        assert update.id == update_id
        assert isinstance(update.thread, LiveThread)
        assert update.thread.id == thread_id

        update = LiveUpdate(reddit, _data=data)
        assert isinstance(update, LiveUpdate)
        assert update.id == update_id
        assert update._fetched

    def test_construct_failure(self, reddit):
        message = "Either 'thread_id' and 'update_id', or '_data' must be provided."
        thread_id = "dummy_thread_id"
        update_id = "dummy_update_id"

        with pytest.raises(TypeError) as excinfo:
            LiveUpdate(reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            LiveUpdate(reddit, thread_id=thread_id)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            LiveUpdate(reddit, update_id=update_id)
        assert str(excinfo.value) == message

    def test_contrib(self, reddit):
        thread_id = "dummy_thread_id"
        update_id = "dummy_update_id"
        update = LiveUpdate(reddit, thread_id, update_id)
        assert isinstance(update.contrib, LiveUpdateContribution)

        data = {"id": "dummy_update_id", "author": "dummy_author"}
        update = LiveUpdate(reddit, _data=data)
        assert isinstance(update.contrib, LiveUpdateContribution)

    def test_setattr(self, reddit):
        data = {"id": "dummy_update_id", "author": "dummy_author"}
        update = LiveUpdate(reddit, _data=data)
        assert isinstance(update.author, Redditor)

    def test_thread(self, reddit):
        thread_id = "dummy_thread_id"
        update_id = "dummy_update_id"

        update = LiveUpdate(reddit, thread_id=thread_id, update_id=update_id)
        assert isinstance(update.thread, LiveThread)
        assert update.thread.id == thread_id



from asyncpraw.models import Subreddit, WikiPage

from ... import UnitTest


class TestWikiPage(UnitTest):
    def test_equality(self, reddit):
        page1 = WikiPage(reddit, subreddit=Subreddit(reddit, "a"), name="x")
        page2 = WikiPage(reddit, subreddit=Subreddit(reddit, "a"), name="2")
        page3 = WikiPage(reddit, subreddit=Subreddit(reddit, "b"), name="1")
        page4 = WikiPage(reddit, subreddit=Subreddit(reddit, "A"), name="x")
        page5 = WikiPage(reddit, subreddit=Subreddit(reddit, "a"), name="X")
        assert page1 == page1
        assert page2 == page2
        assert page3 == page3
        assert page1 != page2
        assert page1 != page3
        assert page1 == page4
        assert page1 == page5

    def test_hash(self, reddit):
        page1 = WikiPage(reddit, subreddit=Subreddit(reddit, "a"), name="x")
        page2 = WikiPage(reddit, subreddit=Subreddit(reddit, "a"), name="2")
        page3 = WikiPage(reddit, subreddit=Subreddit(reddit, "b"), name="1")
        page4 = WikiPage(reddit, subreddit=Subreddit(reddit, "A"), name="x")
        page5 = WikiPage(reddit, subreddit=Subreddit(reddit, "a"), name="X")
        assert hash(page1) == hash(page1)
        assert hash(page2) == hash(page2)
        assert hash(page3) == hash(page3)
        assert hash(page1) != hash(page2)
        assert hash(page1) != hash(page3)
        assert hash(page1) == hash(page4)
        assert hash(page1) == hash(page5)

    def test_repr(self, reddit):
        page = WikiPage(reddit, subreddit=Subreddit(reddit, "a"), name="x")
        assert repr(page) == "WikiPage(subreddit=Subreddit(display_name='a'), name='x')"

    def test_str(self, reddit):
        page = WikiPage(reddit, subreddit=Subreddit(reddit, "a"), name="x")
        assert str(page) == "a/x"



"""Test asyncpraw.models.listing.listing."""
from asyncpraw.models.listing.listing import ModmailConversationsListing, ModNoteListing

from ... import UnitTest


class TestModmailConversationsListing(UnitTest):
    def test_empty_conversations_list(self, reddit):
        assert (
            ModmailConversationsListing(reddit, _data={"conversations": []}).after
            is None
        )


class TestModNoteListing(UnitTest):
    def test_has_next_page(self, reddit):
        assert (
            ModNoteListing(
                reddit, _data={"has_next_page": True, "end_cursor": "end_cursor"}
            ).after
            == "end_cursor"
        )
